/**
 * TrueReality - N64/memory.c
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/*****************************************************************************/
/* include section                                                           */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <config.h>
#include <parser_extern.h>
#include <file_loader_extern.h>
#include <dispatch.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <ansi_esc_codes.h>

#include "registers.h"
#include "rsp_registers.h"
#include "rdp.h"
#include "exceptions.h"
#include "interrupts.h"
#include "dma.h"
#include "memory.h"

#ifdef DEBUG
        #include <debug.h>
#endif

#ifdef SHM
#       include <sys/types.h>
#       include <sys/ipc.h>
#       include <sys/shm.h>
#       include <X11/Xlib.h>
#       include <X11/extensions/XShm.h>
#endif



/*****************************************************************************/
/* define section                                                            */

        /* for mem_todo */
#define DO_NOTHING              0x0000
#define DO_DISPLAY_REFRESH      0x0001
#define DO_RUN_RDP              0x0002
#define UPDATE_MI_REG           0x0004
#define UPDATE_VI_REG           0x0008
#define UPDATE_PI_REG           0x0010
#define UPDATE_SP_REG           0x0020
#define UPDATE_DP_REG           0x0040
#ifdef PUR_VI_INTERRUPT
        #define CHECK_VI_REG            0x0080
#endif
#define CHECK_AI_REG            0x0100
#define PI_DMA                  0x0200
#define PIF2DRAM_DMA            0x0400
#define DRAM2PIF_DMA            0x0800
#define IDMEM2DRAM_DMA          0x1000
#define DRAM2IDMEM_DMA          0x2000

        /* for making code nicer */
#ifdef DEBUG
        #define print_debug_msg(dbg_func, r_w)                                                  \
                if( (dbg_func) )                                                                \
                {                                                                               \
                        printf("DEBUG MEM: [pc]=%08lx [count]=%08lx vaddr=%08lx (%s)\n",        \
                               reg.pc,                                                          \
                               (uWORD)reg.cpr[0][9],                                            \
                               virtualaddr,                                                     \
                               r_w);                                                            \
                }
#else
        #define print_debug_msg(dbg_func, r_w)
#endif

        /* for speed reasons ( inline :) ) */
#ifdef PUR_VI_INTERRUPT
#define final_check()                           \
        if(mem_todo)                            \
        {                                       \
                if(mem_todo & DO_DISPLAY_REFRESH) \
                        do_display_refresh();   \
                if(mem_todo & DO_RUN_RDP)       \
                        rdp_go();               \
                if(mem_todo & UPDATE_VI_REG)    \
                        update_vi_reg();        \
                if(mem_todo & CHECK_VI_REG)     \
                        check_vi_reg();         \
                if(mem_todo & CHECK_AI_REG)     \
                        check_ai_reg();         \
                if(mem_todo & UPDATE_MI_REG)    \
                        update_mi_reg();        \
                if(mem_todo & UPDATE_PI_REG)    \
                        update_pi_reg();        \
                if(mem_todo & UPDATE_SP_REG)    \
                        update_sp_reg();        \
                if(mem_todo & UPDATE_DP_REG)    \
                        update_dp_reg();        \
                if(mem_todo & PI_DMA)           \
                        dma_pi_copy();          \
                if(mem_todo & PIF2DRAM_DMA)     \
                        dma_pif2dram();         \
                if(mem_todo & DRAM2PIF_DMA)     \
                        dma_dram2pif();         \
                if(mem_todo & IDMEM2DRAM_DMA)   \
                        dma_idmem2dram();       \
                if(mem_todo & DRAM2IDMEM_DMA)   \
                        dma_dram2idmem();       \
        }                                       \
        mem_todo = DO_NOTHING;
#endif /* #ifdef PUR_VI_INTERRUPT */

#ifndef PUR_VI_INTERRUPT
#define final_check()                           \
        if(mem_todo)                            \
        {                                       \
                if(mem_todo & DO_DISPLAY_REFRESH) \
                        do_display_refresh();   \
                if(mem_todo & DO_RUN_RDP)       \
                        rdp_go();               \
                if(mem_todo & UPDATE_VI_REG)    \
                        update_vi_reg();        \
                if(mem_todo & CHECK_AI_REG)     \
                        check_ai_reg();         \
                if(mem_todo & UPDATE_MI_REG)    \
                        update_mi_reg();        \
                if(mem_todo & UPDATE_PI_REG)    \
                        update_pi_reg();        \
                if(mem_todo & UPDATE_SP_REG)    \
                        update_sp_reg();        \
                if(mem_todo & UPDATE_DP_REG)    \
                        update_dp_reg();        \
                if(mem_todo & PI_DMA)           \
                        dma_pi_copy();          \
                if(mem_todo & PIF2DRAM_DMA)     \
                        dma_pif2dram();         \
                if(mem_todo & DRAM2PIF_DMA)     \
                        dma_dram2pif();         \
                if(mem_todo & IDMEM2DRAM_DMA)   \
                        dma_idmem2dram();       \
                if(mem_todo & DRAM2IDMEM_DMA)   \
                        dma_dram2idmem();       \
        }                                       \
        mem_todo = DO_NOTHING;
#endif /* #ifdef PUR_VI_INTERRUPT */



/*****************************************************************************/
/* extern section                                                            */

struct  tr_mem  mem;
#ifdef SHM
        struct tr_shminfo shminfo;
#endif





/*****************************************************************************/
/* static section                                                            */

        /* vars */
static  int     mem_todo;
#ifdef PUR_VI_INTERRUPT
                        /* for check_vi_reg */
        static  uWORD   old_vi_status_reg;
        static  uWORD   old_vi_width_reg;
#endif
                /* for audio interface */
static  sWORD   ai_addr_len_buffer_full = 0;

/* the followinng define is not in the define section because it is
   VERY important for the next 3 static arrays */

#define N_SEGMENTS 20

static size_t mem_size[N_SEGMENTS] =
{
        4 * 1024 * 1024,        /* rd_ram   */  /*  0 */
        0x1000,                 /* sp_dmem  */  /*  1 */
        0x1000,                 /* sp_imem  */  /*  2 */
        0x7c0,                  /* pi_rom   */  /*  3 */
        0x40,                   /* pi_ram   */  /*  4 */
        0x40,                   /* pi_ram_w */  /*  5 */
        
        0x28,                   /* rd_reg   */  /*  6 */
        0x20,                   /* sp_reg   */  /*  7 */
        0x20,                   /* dpc_reg  */  /*  8 */
        0x10,                   /* dps_reg  */  /*  9 */
        0x10,                   /* mi_reg   */  /* 10 */
        0x10,                   /* mi_reg_w */  /* 11 */
        0x38,                   /* vi_reg   */  /* 12 */
        0x18,                   /* ai_reg   */  /* 13 */
        0x34,                   /* pi_reg   */  /* 14 */
        0x34,                   /* pi_reg_w */  /* 15 */
        0x20,                   /* ri_reg   */  /* 16 */
        0x1c,                   /* si_reg   */  /* 17 */
        4096,                   /* tmem     */  /* 18 */
        1024                    /* null_mem */  /* 19 */
};

static char *mem_addr[N_SEGMENTS];

#ifdef SHM
        static XShmSegmentInfo mem_shminfo[N_SEGMENTS];
#endif /* SHM */






/*****************************************************************************\
*                                                                             *
* Prototypes                                                                  *
*                                                                             *
\*****************************************************************************/

/* the fkt pointers */
static char * ((**phys_read_fast)(uWORD addr));
static char * ((**phys_write_fast)(uWORD addr));

/* init of the above fkt pointers */
static void init_fast_mem();

/* if there is no real mem these fkt return a pointer to mem.null_mem.
   on a real n64 you get sthg like electrical noise.
   quor: thanks for this info! - niki */
static char *get_electricalnoise(uWORD virtualaddr);
static char *put_electricalnoise(uWORD virtualaddr);

/* these are for TLB mapped mem access.
   ASID is currently not implemented - i think it is not used. */
static char *get_mapped(uWORD virtualaddr);          /* TLB */
static char *put_mapped(uWORD virtualaddr);          /* TLB */

/* here we get the rom data */
static char *get_rom(uWORD virtualaddr);             /* cartridge domains */

/* the rest of the memory */
static char *get_mem_8000_803f(uWORD virtualaddr);   /* rd_mem */
static char *get_mem_83f0_83f0(uWORD virtualaddr);   /* rd_reg */
static char *get_mem_8400_8400(uWORD virtualaddr);   /* dmem/imem */
static char *get_mem_8404_8404(uWORD virtualaddr);   /* sp_reg */
static char *get_mem_8408_8408(uWORD virtualaddr);   /* sp pc/ibist */
static char *get_mem_8410_841f(uWORD virtualaddr);   /* dpc_reg */
static char *get_mem_8420_842f(uWORD virtualaddr);   /* dps_reg */
static char *get_mem_8430_843f(uWORD virtualaddr);   /* mi_reg */
static char *get_mem_8440_844f(uWORD virtualaddr);   /* vi_reg */
static char *get_mem_8450_845f(uWORD virtualaddr);   /* ai_reg */
static char *get_mem_8460_846f(uWORD virtualaddr);   /* pi_reg */
static char *get_mem_8470_847f(uWORD virtualaddr);   /* ri_reg */
static char *get_mem_8480_848f(uWORD virtualaddr);   /* si_reg */
static char *get_mem_9fc0_9fcf(uWORD virtualaddr);   /* pif ram/rom */

static char *put_mem_8000_803f(uWORD virtualaddr);   /* rd_mem */
static char *put_mem_83f0_83f0(uWORD virtualaddr);   /* rd_reg */
static char *put_mem_8400_8400(uWORD virtualaddr);   /* dmem/imem */
static char *put_mem_8404_8404(uWORD virtualaddr);   /* sp_reg */
static char *put_mem_8408_8408(uWORD virtualaddr);   /* sp pc/ibist */
static char *put_mem_8410_841f(uWORD virtualaddr);   /* dpc_reg */
static char *put_mem_8420_842f(uWORD virtualaddr);   /* dps_reg */
static char *put_mem_8430_843f(uWORD virtualaddr);   /* mi_reg */
static char *put_mem_8440_844f(uWORD virtualaddr);   /* vi_reg */
static char *put_mem_8450_845f(uWORD virtualaddr);   /* ai_reg */
static char *put_mem_8460_846f(uWORD virtualaddr);   /* pi_reg */
static char *put_mem_8470_847f(uWORD virtualaddr);   /* ri_reg */
static char *put_mem_8480_848f(uWORD virtualaddr);   /* si_reg */
static char *put_mem_9fc0_9fcf(uWORD virtualaddr);   /* pif ram/rom */

/* used by alloc_n64_mem() */
static  void    write_mem_addr();

/* used by final_check() (after each write to memory) */
static  void    do_display_refresh();
static  void    update_mi_reg();
static  void    update_vi_reg();
static  void    update_pi_reg();
static  void    update_sp_reg();
static  void    update_dp_reg();
static  void    check_ai_reg();
#ifdef PUR_VI_INTERRUPT
static  void    check_vi_reg();
#endif










/*****************************************************************************\
*                                                                             *
* Alloc/init/free the memory used by the emulation (public routines).         *
*                                                                             *
\*****************************************************************************/

int alloc_n64_mem()
{
        int i;
#ifdef SHM
        struct shmid_ds s;
#endif
        
        
        
        fputs("Allocating memory ...", stdout); fflush(stdout);

        for(i=0; i<N_SEGMENTS; i++)
        {
#ifdef SHM
                if(prefs.UseSHM)
                {
                        mem_shminfo[i].shmid = shmget(IPC_PRIVATE, mem_size[i], IPC_CREAT | 0777);
                        if(mem_shminfo[i].shmid < 0) 
                        { 
                                printf("alloc_n64_mem: Couldn't get SHM ID [%d]\n\n", i);
                                return(-1); 
                        }

                        mem_addr[i] = mem_shminfo[i].shmaddr = shmat(mem_shminfo[i].shmid, 0, 0);
                        if( ! mem_addr[i] ) 
                        { 
                                printf("alloc_n64_mem: Couldn't attach shared memory [%d]\n\n", i);
                                return(-1);
                        }

                        mem_shminfo[i].readOnly = False;

                        shmctl(mem_shminfo[i].shmid, IPC_RMID, &s);   /* if someone kills me, shm should be freed */

                        memset(mem_addr[i], 0, mem_size[i]);

                } /* if(prefs.UseSHM) */
                else
#endif
                {
                        mem_addr[i] = (uBYTE *) malloc(mem_size[i]);
                        if( ! mem_addr[i] )
                        {
                                printf("alloc_n64_mem: Not enough memory [%d]\n\n", i);
                                return(-1);
                        }
                        memset(mem_addr[i], 0, mem_size[i]);
                }

        } /* for(i=0; i<N_SEGMENTS; i++) */



        write_mem_addr();



        /* this need to be initialized */

        ((uWORD *)mem.mi_reg)[1] = 0x01010101;
        mem.sp_reg_pc = 0x04001000;

    /* we need 16k entries for read and write */
        phys_read_fast = malloc(0x4000 * sizeof(char * (**)(uWORD)));
        phys_write_fast = malloc(0x4000 * sizeof(char * (**)(uWORD)));
        init_fast_mem();

        puts(" OK\n"
             "-------------------------------------------------------------------------------"); fflush(stdout);

        return(0);

} /* int alloc_n64_mem() */





void clear_n64_mem()
{
        int i;



        fputs("Clearing memory ...", stdout); fflush(stdout);

        for(i=0; i<N_SEGMENTS; i++)
        {
                memset(mem_addr[i], 0, mem_size[i]);
        
        } /* for(i=0; mem_size[i] > 0; i++) */



        /* this need to be initialized */
        
        ((uWORD *)mem.mi_reg)[1] = 0x01010101;
        mem.sp_reg_pc = 0x04001000;



        /* we don't have to do anything in this stage */
        
        mem_todo = DO_NOTHING;


        
#ifdef PUR_VI_INTERRUPT
        /* clr static vi (video interface) dependend vars */
        
        old_vi_status_reg = 0;  
        old_vi_width_reg  = 0;
#endif



        puts(" OK\n"
             "-------------------------------------------------------------------------------"); fflush(stdout);

} /* void clear_n64_mem() */





void free_n64_mem()
{
        int i;
        

        
        fputs("Freeing memory ...", stdout); fflush(stdout);

        for(i=0; i<N_SEGMENTS; i++)
        {
#ifdef SHM
                if(prefs.UseSHM)
                {
                        if(mem_shminfo[i].shmaddr)
                                shmdt(mem_shminfo[i].shmaddr);
                } /* if(prefs.UseSHM) */
                else
#endif
                {
                        if(mem_addr[i])
                                free(mem_addr[i]);
                }

        } /* for(i=0; i<N_SEGMENTS; i++) */

        free(phys_read_fast);
        free(phys_write_fast);

        puts(" OK\n"
             "-------------------------------------------------------------------------------"); fflush(stdout);

} /* void free_n64_mem() */










/*****************************************************************************\
*                                                                             *
* The public memory access routines follow.                                   *
*                                                                             *
\*****************************************************************************/

#ifdef MACOS
/* Gil: do you still need this? - niki */

uWORD *doGetMemPtr(uWORD where)
{
        return(((uWORD *)(phys_read_fast)[where>>18](where)));

} /* uWORD* doGetMemPtr(uWORD where) */
#endif





uBYTE doReadMemByte(uWORD where)
{
#ifdef ENDIAN_IS_LITTLE
        where ^= 0x03;
#endif

        return(*((uBYTE *)(phys_read_fast)[where>>18](where)));

} /* uBYTE doReadMemByte(uWORD where) */





uHWORD doReadMemHalfWord(uWORD where)
{
#ifdef ENDIAN_IS_LITTLE
        where ^= 0x02;
#endif

        return(*((uHWORD *)(phys_read_fast)[where>>18](where)));

} /* uHWORD doReadMemHalfWord(uWORD where) */





uWORD doReadMemWord(uWORD where)
{
        return(*((uWORD *)(phys_read_fast)[where>>18](where)));

} /* uWORD doReadMemWord(uWORD where) */





uDWORD doReadMemDoubleWord(uWORD where)
{
#ifdef ENDIAN_IS_LITTLE
        uDWORD value;
#endif



#ifdef ENDIAN_IS_BIG
        return(*((uDWORD *)(phys_read_fast)[where>>18](where)));
#endif
#ifdef ENDIAN_IS_LITTLE
        value = *((uDWORD *)(phys_read_fast)[where>>18](where));
        return( (value >> 32) | (value << 32) );
#endif

} /* uDWORD doReadMemDoubleWord(uWORD where) */





void doWriteMemByte(uBYTE what, uWORD where)
{
#ifdef ENDIAN_IS_LITTLE
        where ^= 0x03;
#endif

        *((uBYTE *)(phys_write_fast)[where>>18](where)) = what;

        final_check();

} /* void doWriteMemByte(uBYTE what, uWORD where) */





void doWriteMemHalfWord(uHWORD what, uWORD where)
{
#ifdef ENDIAN_IS_LITTLE
        where ^= 0x02;
#endif

        *((uHWORD *)(phys_write_fast)[where>>18](where)) = what;

        final_check();

} /* void doWriteMemHalfWord(uHWORD what, uWORD where) */





void doWriteMemWord(uWORD what, uWORD where)
{
        *((uWORD *)(phys_write_fast)[where>>18](where)) = what;

        final_check();

} /* void doWriteMemWord(uWORD what, uWORD where) */





void doWriteMemDoubleWord(uDWORD what, uWORD where)
{
#ifdef ENDIAN_IS_BIG
        *((uDWORD *)(phys_write_fast)[where>>18](where)) = what;
#endif
#ifdef ENDIAN_IS_LITTLE
        *((uDWORD *)(phys_write_fast)[where>>18](where)) = (what << 32) | (what >> 32);
#endif

        final_check();

} /* void doWriteMemDoubleWord(uDWORD what, uWORD where) */










/*****************************************************************************\
*                                                                             *
* The static fast memory access routines follow.                              *
* Including the init routine and all function pointer routines.               *
*                                                                             *
* I know that this section is VERY huge!!! But there is no serious way        *
* to do it not like this.                                                     *
*                                                                             *
\*****************************************************************************/

/**
 *
 * initialisation of memory function pointers
 * - called by alloc_n64_mem()
 *
**/

static void init_fast_mem()
{
        int i;



    /* the '>>2' in the following for() loops are just used because i
       wanted to use nice values before the '>>2' - niki */

    /**
     *
     * 0x00000000 - 0x7fffffff: mapped memory
     * 0x80000000 - 0x9fffffff: physical memory (cachable)
     * 0xa0000000 - 0xbfffffff: physical memory (uncachable) (copy of the above!)
     * 0xc0000000 - 0xdfffffff: mapped memory
     * 0xe0000000 - 0xffffffff: mapped memory
     *
     * oh - i forgot to tell: we don't care if the mem is cached or uncached :-)
     *
    **/



    /**
     * 0x00000000 - 0x7fffffff: mapped memory
    **/

    /* mapped */
        for( i = (0x0000>>2); i <= (0x7fff>>2); i++ )
        {
                phys_read_fast[i] = get_mapped;
                phys_write_fast[i] = put_mapped;
        }

    /* we map the first 4M (rdram) of mem to the first 4M of virtual
       addresses. THIS IS A HACK! - niki */

    /* as i thought before - we dont need this hack - niki */

    /* seems to be incorrect - niki
        for( i = (0x0000>>2); i <= (0x003f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8000_803f;
                phys_write_fast[i] = put_mem_8000_803f;
        }
    */


    /**
     * 0x80000000 - 0x9fffffff: physical memory (cachable)
    **/

    /* rdram */
        for( i = (0x8000>>2); i <= (0x803f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8000_803f;
                phys_write_fast[i] = put_mem_8000_803f;
        }

    /* not used */
        for( i = (0x8040>>2); i <= (0x83ef>>2); i++ )
        {
                phys_read_fast[i] = get_electricalnoise;
                phys_write_fast[i] = put_electricalnoise;
        }

    /* rdram reg */
        for( i = (0x83f0>>2); i <= (0x83f0>>2); i++ )
        {
                phys_read_fast[i] = get_mem_83f0_83f0;
                phys_write_fast[i] = put_mem_83f0_83f0;
        }

    /* not used */
        for( i = (0x83f4>>2); i <= (0x83ff>>2); i++ )
        {
                phys_read_fast[i] = get_electricalnoise;
                phys_write_fast[i] = put_electricalnoise;
        }

    /* dmem/imem */
        for( i = (0x8400>>2); i <= (0x8400>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8400_8400;
                phys_write_fast[i] = put_mem_8400_8400;
        }

    /* sp reg */
        for( i = (0x8404>>2); i <= (0x8404>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8404_8404;
                phys_write_fast[i] = put_mem_8404_8404;
        }

    /* sp pc/ibist */
        for( i = (0x8408>>2); i <= (0x8408>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8408_8408;
                phys_write_fast[i] = put_mem_8408_8408;
        }

    /* not used */
        for( i = (0x840c>>2); i <= (0x840f>>2); i++ )
        {
                phys_read_fast[i] = get_electricalnoise;
                phys_write_fast[i] = put_electricalnoise;
        }

    /* dpc reg */
        for( i = (0x8410>>2); i <= (0x841f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8410_841f;
                phys_write_fast[i] = put_mem_8410_841f;
        }

    /* dps reg */
        for( i = (0x8420>>2); i <= (0x824f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8420_842f;
                phys_write_fast[i] = put_mem_8420_842f;
        }

    /* mi reg */
        for( i = (0x8430>>2); i <= (0x843f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8430_843f;
                phys_write_fast[i] = put_mem_8430_843f;
        }

    /* vi reg */
        for( i = (0x8440>>2); i <= (0x844f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8440_844f;
                phys_write_fast[i] = put_mem_8440_844f;
        }

    /* ai reg */
        for( i = (0x8450>>2); i <= (0x845f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8450_845f;
                phys_write_fast[i] = put_mem_8450_845f;
        }

    /* pi reg */
        for( i = (0x8460>>2); i <= (0x846f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8460_846f;
                phys_write_fast[i] = put_mem_8460_846f;
        }

    /* ri reg */
        for( i = (0x8470>>2); i <= (0x847f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8470_847f;
                phys_write_fast[i] = put_mem_8470_847f;
        }

    /* si reg */
        for( i = (0x8480>>2); i <= (0x848f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8480_848f;
                phys_write_fast[i] = put_mem_8480_848f;
        }

    /* not used */
        for( i = (0x8490>>2); i <= (0x849f>>2); i++ )
        {
                phys_read_fast[i] = get_electricalnoise;
                phys_write_fast[i] = put_electricalnoise;
        }

    /* cartridge domain 2 addr 1 */
        for( i = (0x8500>>2); i <= (0x85ff>>2); i++ )
        {
            /* pong seems not to work with this!
                   phys_read_fast[i] = get_rom;
            */
                phys_read_fast[i] = get_electricalnoise;
                phys_write_fast[i] = put_electricalnoise;   /* rom is not writable :) */
        }

    /* cartridge domain 1 addr 1 */
        for( i = (0x8600>>2); i <= (0x87ff>>2); i++ )
        {
                phys_read_fast[i] = get_rom;
                phys_write_fast[i] = put_electricalnoise;   /* rom is not writable :) */
        }

    /* cartridge domain 2 addr 2 */
        for( i = (0x8800>>2); i <= (0x8fff>>2); i++ )
        {
                phys_read_fast[i] = get_rom;
                phys_write_fast[i] = put_electricalnoise;   /* rom is not writable :) */
        }

    /* cartridge domain 1 addr 2 */
        for( i = (0x9000>>2); i <= (0x9fbf>>2); i++ )
        {
                phys_read_fast[i] = get_rom;
                phys_write_fast[i] = put_electricalnoise;   /* rom is not writable :) */
        }

    /* pif ram/rom */
        for( i = (0x9fc0>>2); i <= (0x9fcf>>2); i++ )
        {
                phys_read_fast[i] = get_mem_9fc0_9fcf;
                phys_write_fast[i] = put_mem_9fc0_9fcf;
        }

    /* cartridge domain 1 addr 3 */
        for( i = (0x9fc0>>2); i <= (0x9fff>>2); i++ )
        {
                phys_read_fast[i] = get_rom;
                phys_write_fast[i] = put_electricalnoise;   /* rom is not writable :) */
        }



    /**
     * 0xa0000000 - 0xbfffffff: physical memory (uncachable) (copy of the above!)
    **/

    /* rdram */
        for( i = (0xa000>>2); i <= (0xa03f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8000_803f;
                phys_write_fast[i] = put_mem_8000_803f;
        }

    /* not used */
        for( i = (0xa040>>2); i <= (0xa3ef>>2); i++ )
        {
                phys_read_fast[i] = get_electricalnoise;
                phys_write_fast[i] = put_electricalnoise;
        }

    /* rdram reg */
        for( i = (0xa3f0>>2); i <= (0xa3f0>>2); i++ )
        {
                phys_read_fast[i] = get_mem_83f0_83f0;
                phys_write_fast[i] = put_mem_83f0_83f0;
        }

    /* not used */
        for( i = (0xa3f4>>2); i <= (0xa3ff>>2); i++ )
        {
                phys_read_fast[i] = get_electricalnoise;
                phys_write_fast[i] = put_electricalnoise;
        }

    /* dmem/imem */
        for( i = (0xa400>>2); i <= (0xa400>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8400_8400;
                phys_write_fast[i] = put_mem_8400_8400;
        }

    /* sp reg */
        for( i = (0xa404>>2); i <= (0xa404>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8404_8404;
                phys_write_fast[i] = put_mem_8404_8404;
        }

    /* sp pc/ibist */
        for( i = (0xa408>>2); i <= (0xa408>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8408_8408;
                phys_write_fast[i] = put_mem_8408_8408;
        }

    /* not used */
        for( i = (0xa40c>>2); i <= (0xa40f>>2); i++ )
        {
                phys_read_fast[i] = get_electricalnoise;
                phys_write_fast[i] = put_electricalnoise;
        }

    /* dpc reg */
        for( i = (0xa410>>2); i <= (0xa41f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8410_841f;
                phys_write_fast[i] = put_mem_8410_841f;
        }

    /* dps reg */
        for( i = (0xa420>>2); i <= (0xa24f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8420_842f;
                phys_write_fast[i] = put_mem_8420_842f;
        }

    /* mi reg */
        for( i = (0xa430>>2); i <= (0xa43f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8430_843f;
                phys_write_fast[i] = put_mem_8430_843f;
        }

    /* vi reg */
        for( i = (0xa440>>2); i <= (0xa44f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8440_844f;
                phys_write_fast[i] = put_mem_8440_844f;
        }

    /* ai reg */
        for( i = (0xa450>>2); i <= (0xa45f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8450_845f;
                phys_write_fast[i] = put_mem_8450_845f;
        }

    /* pi reg */
        for( i = (0xa460>>2); i <= (0xa46f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8460_846f;
                phys_write_fast[i] = put_mem_8460_846f;
        }

    /* ri reg */
        for( i = (0xa470>>2); i <= (0xa47f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8470_847f;
                phys_write_fast[i] = put_mem_8470_847f;
        }

    /* si reg */
        for( i = (0xa480>>2); i <= (0xa48f>>2); i++ )
        {
                phys_read_fast[i] = get_mem_8480_848f;
                phys_write_fast[i] = put_mem_8480_848f;
        }

    /* not used */
        for( i = (0xa490>>2); i <= (0xa49f>>2); i++ )
        {
                phys_read_fast[i] = get_electricalnoise;
                phys_write_fast[i] = put_electricalnoise;
        }

    /* cartridge domain 2 addr 1 */
        for( i = (0xa500>>2); i <= (0xa5ff>>2); i++ )
        {
                phys_read_fast[i] = get_rom;
                phys_write_fast[i] = put_electricalnoise;   /* rom is not writable :) */
        }

    /* cartridge domain 1 addr 1 */
        for( i = (0xa600>>2); i <= (0xa7ff>>2); i++ )
        {
                phys_read_fast[i] = get_rom;
                phys_write_fast[i] = put_electricalnoise;   /* rom is not writable :) */
        }

    /* cartridge domain 2 addr 2 */
        for( i = (0xa800>>2); i <= (0xafff>>2); i++ )
        {
                phys_read_fast[i] = get_rom;
                phys_write_fast[i] = put_electricalnoise;   /* rom is not writable :) */
        }

    /* cartridge domain 1 addr 2 */
        for( i = (0xb000>>2); i <= (0xbfbf>>2); i++ )
        {
                phys_read_fast[i] = get_rom;
                phys_write_fast[i] = put_electricalnoise;   /* rom is not writable :) */
        }

    /* pif ram/rom */
        for( i = (0xbfc0>>2); i <= (0xbfcf>>2); i++ )
        {
                phys_read_fast[i] = get_mem_9fc0_9fcf;
                phys_write_fast[i] = put_mem_9fc0_9fcf;
        }

    /* cartridge domain 1 addr 3 */
        for( i = (0xbfd0>>2); i <= (0xbfff>>2); i++ )
        {
                phys_read_fast[i] = get_rom;
                phys_write_fast[i] = put_electricalnoise;   /* rom is not writable :) */
        }



    /**
     * 0xc0000000 - 0xdfffffff: mapped memory
    **/

    /* mapped */
        for( i = (0xc000>>2); i <= (0xdfff>>2); i++ )
        {
                phys_read_fast[i] = get_mapped;
                phys_write_fast[i] = put_mapped;
        }



    /**
     * 0xe0000000 - 0xffffffff: mapped memory
    **/

    /* mapped */
        for( i = (0xe000>>2); i <= (0xffff>>2); i++ )
        {
                phys_read_fast[i] = get_mapped;
                phys_write_fast[i] = put_mapped;
        }


} /* static void init_fast_mem() */





/**
 *
 * now the routines which the function pointer point to ;-)
 *
**/

static char *get_electricalnoise(uWORD virtualaddr)
{
        print_debug_msg(1, "read  - noise (no mem here!)");
        return(mem.null_mem);
}

static char *put_electricalnoise(uWORD virtualaddr)
{
        print_debug_msg(1, "write - noise (no mem here!)");
        return(&(mem.null_mem[4]));
}

static char *get_mapped(uWORD virtualaddr)          /* TLB */
{
        int                i;
        struct rs4300i_tlb *tlb;



        //printf("TLB: read from: %08lx\n", virtualaddr);

        for(i=0; i<32; i++) /* we search from tlb entry 0 up to 31 */
        {
                tlb = &(reg.tlb[i]);

                if( (virtualaddr & tlb->invmask) == (tlb->hl & tlb->invmask) )
                {
                    /* we checked the upper bits and found a tlb match */

                        if(tlb->g) /* global bit */
                        {
                                if(virtualaddr & tlb->even_odd_checkbit) /* odd entry (EntryLo1) */
                                {
                                        if(tlb->lh & 0x02) /* valid bit */
                                        {
                                                uWORD addr = 0x80000000
                                                        | ((tlb->lh << 6) & (tlb->invmask >> 1))
                                                        | (virtualaddr & (tlb->mask >> 1));

                                                //printf("physical addr: %08lx\n", addr);
                                                return( phys_read_fast[addr>>18](addr) ); /* recursive! */
                                        }
                                        else
                                        {
                                            /* tlb invalid exception */
                                                reg.cpr[0][BADVADDR] = virtualaddr;
                                                tlb_load_or_store = EXC_TLB_LOAD;
                                                reg.do_or_check_sthg = RS4300I_DO_TLB_INVALID_EXCEPTION;
                                                return(get_electricalnoise(virtualaddr));
                                        }
                                }
                                else /* even entry (EntryLo0) */
                                {
                                        if(tlb->ll & 0x02) /* valid bit */
                                        {
                                                uWORD addr = 0x80000000
                                                        | ((tlb->ll << 6) & (tlb->invmask >> 1))
                                                        | (virtualaddr & (tlb->mask >> 1));

                                                //printf("physical addr: %08lx\n", addr);
                                                return( phys_read_fast[addr>>18](addr) ); /* recursive! */
                                        }
                                        else
                                        {
                                            /* tlb invalid exception */
                                                reg.cpr[0][BADVADDR] = virtualaddr;
                                                tlb_load_or_store = EXC_TLB_LOAD;
                                                reg.do_or_check_sthg = RS4300I_DO_TLB_INVALID_EXCEPTION;
                                                return(get_electricalnoise(virtualaddr));
                                        }
                                }
                        }
                        else /* global bit not set */
                        {
                            /* -> check ASID */
                            /* BUG: we dont check the ASID here - not implemented */

                                reg.cpr[0][BADVADDR] = virtualaddr;
                                tlb_load_or_store = EXC_TLB_LOAD;
                                //reg.do_or_check_sthg = RS4300I_DO_TLB_MISS_EXCEPTION;
                                //reg.do_or_check_sthg = RS4300I_DO_TLB_MISS_EXCEPTION;
                                return(get_electricalnoise(virtualaddr));
                        }
                }
        }

    /* there is no valid tlb entry -> tlb miss exception */
        reg.cpr[0][BADVADDR] = virtualaddr;
        tlb_load_or_store = EXC_TLB_LOAD;
    /* seems not to work - niki */
        //reg.do_or_check_sthg = RS4300I_DO_TLB_MISS_EXCEPTION;
        //reg.do_or_check_sthg = RS4300I_DO_TLB_MISS_EXCEPTION;

        return(get_electricalnoise(virtualaddr));
}


static char *put_mapped(uWORD virtualaddr)          /* TLB */
{
        int                i;
        struct rs4300i_tlb *tlb;



        //printf("TLB: write to: %08lx\n", virtualaddr);

        for(i=0; i<32; i++) /* we search from tlb entry 0 up to 31 */
        {
                tlb = &(reg.tlb[i]);

                if( (virtualaddr & tlb->invmask) == (tlb->hl & tlb->invmask) )
                {
                    /* we checked the upper bits and found a tlb match */

                        if(tlb->g) /* global bit */
                        {
                                if(virtualaddr & tlb->even_odd_checkbit) /* odd entry (EntryLo1) */
                                {
                                        if(tlb->lh & 0x02) /* valid bit */
                                        {
                                                if(tlb->lh & 0x04) /* dirty bit */
                                                {
                                                        uWORD addr = 0x80000000
                                                                | ((tlb->lh << 6) & (tlb->invmask >> 1))
                                                                | (virtualaddr & (tlb->mask >> 1));

                                                        //printf("physical addr: %08lx\n", addr);
                                                        return( phys_read_fast[addr>>18](addr) ); /* recursive! */
                                                }
                                                else
                                                {
                                                    /* tlb modification exception */
                                                        reg.cpr[0][BADVADDR] = virtualaddr;
                                                        tlb_load_or_store = EXC_TLB_STORE;
                                                        reg.do_or_check_sthg = RS4300I_DO_TLB_MODIFICATION_EXCEPTION;
                                                        return(get_electricalnoise(virtualaddr));
                                                }
                                        }
                                        else
                                        {
                                            /* tlb invalid exception */
                                                reg.cpr[0][BADVADDR] = virtualaddr;
                                                tlb_load_or_store = EXC_TLB_STORE;
                                                reg.do_or_check_sthg = RS4300I_DO_TLB_INVALID_EXCEPTION;
                                                return(get_electricalnoise(virtualaddr));
                                        }
                                }
                                else /* even entry (EntryLo0) */
                                {
                                        if(tlb->ll & 0x02) /* valid bit */
                                        {
                                                if(tlb->ll & 0x03) /* dirty bit */
                                                {
                                                        uWORD addr = 0x80000000
                                                                | ((tlb->lh << 6) & (tlb->invmask >> 1))
                                                                | (virtualaddr & (tlb->mask >> 1));

                                                        //printf("physical addr: %08lx\n", addr);
                                                        return( phys_read_fast[addr>>18](addr) ); /* recursive! */
                                                }
                                                else
                                                {
                                                    /* tlb modified exception */
                                                        reg.cpr[0][BADVADDR] = virtualaddr;
                                                        tlb_load_or_store = EXC_TLB_STORE;
                                                        reg.do_or_check_sthg = RS4300I_DO_TLB_MODIFICATION_EXCEPTION;
                                                        return(get_electricalnoise(virtualaddr));
                                                }
                                        }
                                        else
                                        {
                                            /* tlb invalid exception */
                                                reg.cpr[0][BADVADDR] = virtualaddr;
                                                tlb_load_or_store = EXC_TLB_STORE;
                                                reg.do_or_check_sthg = RS4300I_DO_TLB_INVALID_EXCEPTION;
                                                return(get_electricalnoise(virtualaddr));
                                        }
                                }
                        }
                        else /* global bit not set */
                        {
                            /* -> check ASID */
                            /* BUG: we dont check the ASID here - not implemented */

                                reg.cpr[0][BADVADDR] = virtualaddr;
                                tlb_load_or_store = EXC_TLB_STORE;
                                //reg.do_or_check_sthg = RS4300I_DO_TLB_MISS_EXCEPTION;
                                //reg.do_or_check_sthg = RS4300I_DO_TLB_MISS_EXCEPTION;
                                return(get_electricalnoise(virtualaddr));
                        }
                }
        }

    /* there is no valid tlb entry -> tlb miss exception */
        reg.cpr[0][BADVADDR] = virtualaddr;
        tlb_load_or_store = EXC_TLB_STORE;
    /* seems not to work - niki */
        //reg.do_or_check_sthg = RS4300I_DO_TLB_MISS_EXCEPTION;
        //reg.do_or_check_sthg = RS4300I_DO_TLB_MISS_EXCEPTION;

        return(put_electricalnoise(virtualaddr));
}






static char *get_mem_8000_803f(uWORD virtualaddr)   /* rd_mem */
{
        print_debug_msg(prefs.debug & DBG_RD_RAM, "read  - rd ram");
        return(mem.rd_ram + (virtualaddr & 0x3fffff));
}

static char *get_mem_83f0_83f0(uWORD virtualaddr)   /* rd_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x03f00028)
        {
                print_debug_msg(prefs.debug & DBG_RD_REG, "read  - rd reg");
                return(mem.rd_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_8400_8400(uWORD virtualaddr)   /* dmem/imem */
{
        if((virtualaddr & 0x1fffffff) < 0x04001000)
        {
                print_debug_msg(prefs.debug & DBG_SP_DMEM, "read  - sp dmem");
                return(mem.sp_dmem + (virtualaddr & 0xfff));
        }
        else if((virtualaddr & 0x1fffffff) < 0x04002000)
        {
                print_debug_msg(prefs.debug & DBG_SP_IMEM, "read  - sp imem");
                return(mem.sp_imem + (virtualaddr & 0xfff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_8404_8404(uWORD virtualaddr)   /* sp_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04040020)
        {
                print_debug_msg(prefs.debug & DBG_SP_REG, "read  - sp reg");
                return(mem.sp_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_8408_8408(uWORD virtualaddr)   /* sp pc/ibist */
{
        if((virtualaddr & 0x1fffffff) < 0x04080004)
        {
                print_debug_msg(prefs.debug & DBG_SP_PC_REG, "read  - sp pc");
                return(((char *)&(mem.sp_reg_pc)) + (virtualaddr & 0x03));
        }
        else if((virtualaddr & 0x1fffffff) < 0x04080008)
        {
                print_debug_msg(prefs.debug & DBG_SP_IBIST_REG, "read  - sp ibist");
                return(((char *)&(mem.sp_reg_ibist)) + virtualaddr - 0x04080004);
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_8410_841f(uWORD virtualaddr)   /* dpc_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04100020)
        {
                print_debug_msg(prefs.debug & DBG_DPC_REG, "read  - dpc reg");
                return(mem.dpc_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_8420_842f(uWORD virtualaddr)   /* dps_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04200010)
        {
                print_debug_msg(prefs.debug & DBG_DPS_REG, "read  - dps reg");
                return(mem.dps_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_8430_843f(uWORD virtualaddr)   /* mi_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04300010)
        {
                print_debug_msg(prefs.debug & DBG_MI_REG, "read  - mi reg");
                return(mem.mi_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_8440_844f(uWORD virtualaddr)   /* vi_reg */
{
        static sWORD current_verical_line = -1;

        if((virtualaddr & 0x1fffffff) < 0x04400038)
        {
                print_debug_msg(prefs.debug & DBG_VI_REG, "read  - vi reg");
                switch(virtualaddr & 0xff)
                {
                    case 0x10:
                    case 0x11:
                    case 0x12:
                    case 0x13:
                        if(current_verical_line < 512)
                                current_verical_line += 2;
                        else
                                current_verical_line = 0;
                        return(((char *)&current_verical_line) + (virtualaddr & 0x03));
                }
                return(mem.vi_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_8450_845f(uWORD virtualaddr)   /* ai_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04500018)
        {
                print_debug_msg(prefs.debug & DBG_AI_REG, "read  - ai reg");

                switch(virtualaddr & 0xff)
                {
                    case 0x00:   /* AI_DRAM_ADDR_REG */
                    case 0x01:
                    case 0x02:
                    case 0x03:
                        break;
                    
                    case 0x04:   /* AI_LEN_REG */
                    case 0x05:
                    case 0x06:
                    case 0x07:
                        ((uWORD *)mem.ai_reg)[1] = 0; /* should count to zero! */
                        break;
                    
                    case 0x08:   /* AI_CONTROL_REG */
                    case 0x09:
                    case 0x0a:
                    case 0x0b:
                        break;

                    case 0x0c:   /* AI_STATUS_REG */
                    case 0x0d:
                    case 0x0e:
                    case 0x0f:
/*
                        if(dispatch(AUDIO_READY, 0))
                        {
                                puts("AUDIO READY"); fflush(stdout);
                                ((uWORD *)mem.ai_reg)[3] = 0x00000000;
                        }
                        else
                        {
                                puts("AUDIO NOT READY"); fflush(stdout);
                                ((uWORD *)mem.ai_reg)[3] = 0xc0000001;
                        }
*/
                        if(ai_addr_len_buffer_full == 2)
                                ((uWORD *)mem.ai_reg)[3] = 0x80000000;
                        else
                                ((uWORD *)mem.ai_reg)[3] = 0x00000000;

                        if(ai_addr_len_buffer_full == 3)
                                ai_addr_len_buffer_full = 0;

                        /* ??? bit 0 ??? */

                        break;

                    case 0x10:   /* AI_DACRATE_REG */
                    case 0x11:
                    case 0x12:
                    case 0x13:
                        break;

                    case 0x14:   /* AI_BITRATE_REG */
                    case 0x15:
                    case 0x16:
                    case 0x17:
                        break;
                        
                } /* switch(virtualaddr & 0xff) */

                return(mem.ai_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_8460_846f(uWORD virtualaddr)   /* pi_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04600034)
        {
                print_debug_msg(prefs.debug & DBG_PI_REG, "read  - pi reg");
                return(mem.pi_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_8470_847f(uWORD virtualaddr)   /* ri_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04700020)
        {
                print_debug_msg(prefs.debug & DBG_RI_REG, "read  - ri reg");
                return(mem.ri_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_8480_848f(uWORD virtualaddr)   /* si_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x0480001c)
        {
                print_debug_msg(prefs.debug & DBG_SI_REG, "read  - si reg");
                return(mem.si_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_rom(uWORD virtualaddr)             /* cartridge domains */
{
        if((virtualaddr & 0x01ffffff) < rom.length)
        {
                //print_debug_msg(prefs.debug & DBG_PI_DOM2_ADDR1, "read  - rom");
                return(rom.image + (virtualaddr & 0x01ffffff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *get_mem_9fc0_9fcf(uWORD virtualaddr)   /* pif ram/rom */
{
        if((virtualaddr & 0x1fffffff) < 0x1fc007c0)
        {
                print_debug_msg(prefs.debug & DBG_PI_ROM, "read");
                return(mem.pi_rom + (virtualaddr & 0xfff));
        }
        else if((virtualaddr & 0x1fffffff) < 0x1fc00800)
        {
                print_debug_msg(prefs.debug & DBG_PI_RAM, "read");
                return(mem.pi_ram + (virtualaddr & 0x1fffffff) - 0x1fc007c0);
        }
        else
                return(get_electricalnoise(virtualaddr));
}





static char *put_mem_8000_803f(uWORD virtualaddr)   /* rd_mem */
{
        print_debug_msg(prefs.debug & DBG_RD_RAM, "write  - rd ram");
        return(mem.rd_ram + (virtualaddr & 0x3fffff));
}

static char *put_mem_83f0_83f0(uWORD virtualaddr)   /* rd_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x03f00028)
        {
                print_debug_msg(prefs.debug & DBG_RD_REG, "write - rd reg");
                return(mem.rd_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_8400_8400(uWORD virtualaddr)   /* dmem/imem */
{
        if((virtualaddr & 0x1fffffff) < 0x04001000)
        {
                print_debug_msg(prefs.debug & DBG_SP_DMEM, "write - sp dmem");
                return(mem.sp_dmem + (virtualaddr & 0xfff));
        }
        else if((virtualaddr & 0x1fffffff) < 0x04002000)
        {
                print_debug_msg(prefs.debug & DBG_SP_IMEM, "write - sp imem");
                return(mem.sp_imem + (virtualaddr & 0xfff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_8404_8404(uWORD virtualaddr)   /* sp_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04040020)
        {
                print_debug_msg(prefs.debug & DBG_SP_REG, "write - sp reg");

                switch(virtualaddr & 0xff)
                {
                    case 0x00:   /* SP_MEM_ADDR_REG */
                    case 0x01:
                    case 0x02:
                    case 0x03:
                        break;
                    
                    case 0x04:   /* SP_DRAM_ADDR_REG */
                    case 0x05:
                    case 0x06:
                    case 0x07:
                        break;
                    
                    case 0x08:   /* SP_RD_LEN_REG */
                    case 0x09:
                    case 0x0a:
                    case 0x0b:
                        mem_todo |= DRAM2IDMEM_DMA;
                        break;
                        
                    
                    case 0x0c:   /* SP_WR_LEN_REG */
                    case 0x0d:
                    case 0x0e:
                    case 0x0f:
                        mem_todo |= IDMEM2DRAM_DMA;
                        break;
                    
                    case 0x10:   /* SP_STATUS_REG */
                    case 0x11:
                    case 0x12:
                    case 0x13:
                        mem_todo |= UPDATE_SP_REG;
                        return((char *)&(mem.sp_reg_status_w) + (virtualaddr & 0x3));
                    
                    case 0x14:   /* SP_DMA_FULL_REG (ro) */
                    case 0x15:
                    case 0x16:
                    case 0x17:
                        return(mem.null_mem);
                    
                    case 0x18:   /* SP_DMA_BUSY_REG (ro) */
                    case 0x19:
                    case 0x1a:
                    case 0x1b:
                        return(mem.null_mem);
                    
                    case 0x1c:   /* SP_SEMAPHORE_REG */
                    case 0x1d:
                    case 0x1e:
                    case 0x1f:
                        break;
                    
                } /* switch(virtualaddr & 0xff) */

                return(mem.sp_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_8408_8408(uWORD virtualaddr)   /* sp pc/ibist */
{
        if((virtualaddr & 0x1fffffff) < 0x04080004)
        {
                print_debug_msg(prefs.debug & DBG_SP_PC_REG, "write - sp pc");
                return(((char *)&(mem.sp_reg_pc)) + (virtualaddr & 0x03));
        }
        else if((virtualaddr & 0x1fffffff) < 0x04080008)
        {
                print_debug_msg(prefs.debug & DBG_SP_IBIST_REG, "write - sp ibist");
                return(((char *)&(mem.sp_reg_ibist)) + virtualaddr - 0x04080004);
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_8410_841f(uWORD virtualaddr)   /* dpc_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04100020)
        {
                print_debug_msg(prefs.debug & DBG_DPC_REG, "write - dpc reg");

                switch(virtualaddr & 0xff)
                {
                    case 0x00:   /* DPC_START_REG */
                    case 0x01:
                    case 0x02:
                    case 0x03:
                        break;

                    case 0x04:   /* DPC_END_REG */
                    case 0x05:
                    case 0x06:
                    case 0x07:
                        mem_todo |= DO_RUN_RDP;
                        break;

                    case 0x08:   /* DPC_CURRENT_REG (ro) */
                    case 0x09:
                    case 0x0a:
                    case 0x0b:
                        return(mem.null_mem);

                    case 0x0c:   /* DPC_STATUS_REG */
                    case 0x0d:
                    case 0x0e:
                    case 0x0f:
                        mem_todo |= UPDATE_DP_REG;
                        return((char *)&(mem.dpc_reg_status_w) + (virtualaddr & 0x3));

                    case 0x10:   /* DPC_CLOCK_REG (ro) */
                    case 0x11:
                    case 0x12:
                    case 0x13:
                        return(mem.null_mem);

                    case 0x14:   /* DPC_BUFBUSY_REG (ro) */
                    case 0x15:
                    case 0x16:
                    case 0x17:
                        return(mem.null_mem);

                    case 0x18:   /* DPC_PIPEBUSY_REG (ro) */
                    case 0x19:
                    case 0x1a:
                    case 0x1b:
                        return(mem.null_mem);

                    case 0x1c:   /* DPC_TMEM_REG (ro) */
                    case 0x1d:
                    case 0x1e:
                    case 0x1f:
                        return(mem.null_mem);

                } /* switch(virtualaddr & 0xff) */

                return(mem.dpc_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_8420_842f(uWORD virtualaddr)   /* dps_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04200010)
        {
                print_debug_msg(prefs.debug & DBG_DPS_REG, "write - dps reg");
                return(mem.dps_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_8430_843f(uWORD virtualaddr)   /* mi_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04300010)
        {
                print_debug_msg(prefs.debug & DBG_MI_REG, "write - mi reg");

                switch(virtualaddr & 0xf)
                {
                    case 0x00:   /* MI_MODE_REG */
                    case 0x01:
                    case 0x02:
                    case 0x03:
                        mem_todo |= UPDATE_MI_REG;
                        return((char *)&(mem.mi_reg_mode_w) + (virtualaddr & 0x3));
                    
                    case 0x04:   /* MI_VERSION_REG */
                    case 0x05:
                    case 0x06:
                    case 0x07:
                        return(mem.null_mem);
                    
                    case 0x08:   /* MI_INTR_REG */
                    case 0x09:
                    case 0x0a:
                    case 0x0b:
                        return(mem.null_mem);
                        
                    
                    case 0x0c:   /* MI_INTR_MASK_REG */
                    case 0x0d:
                    case 0x0e:
                    case 0x0f:
                        mem_todo |= UPDATE_MI_REG;
                        reg.do_or_check_sthg |= RS4300I_CHECK_MI_INTERRUPTS;
                        return((char *)&(mem.mi_reg_intr_mask_w) + (virtualaddr & 0x3));

                } /* switch(virtualaddr & 0xf) */

                return(mem.mi_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_8440_844f(uWORD virtualaddr)   /* vi_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04400038)
        {
                print_debug_msg(prefs.debug & DBG_VI_REG, "write - vi reg");

                switch(virtualaddr & 0xff)
                {
                    case 0x00:   /* VI_CONTROL_REG */
                    case 0x01:
                    case 0x02:
                    case 0x03:
#ifdef PUR_VI_INTERRUPT
                        mem_todo |= CHECK_VI_REG;
#endif
                        break;
                    
                    case 0x04:   /* VI_ORIGIN_REG */
                    case 0x05:
                    case 0x06:
                    case 0x07:
                        mem_todo |= UPDATE_VI_REG;
#ifdef PUR_VI_INTERRUPT
#if DISPLAY_SUPPORT
                        mem_todo |= DO_DISPLAY_REFRESH;
#endif
#endif
                        break;
                    
                    case 0x08:   /* VI_WIDTH_REG */
                    case 0x09:
                    case 0x0a:
                    case 0x0b:
                        mem_todo |= UPDATE_VI_REG;
#ifdef PUR_VI_INTERRUPT
                        mem_todo |= CHECK_VI_REG;
#endif
                        break;

                    case 0x0c:   /* VI_INTR_REG */
                    case 0x0d:
                    case 0x0e:
                    case 0x0f:
                        break;

                    case 0x10:   /* VI_CURRENT_REG */
                    case 0x11:
                    case 0x12:
                    case 0x13:
#ifndef PUR03A_COMPATIBLE
                        ((uWORD *)mem.mi_reg)[2] &= ~MI_INTR_VI;   /* clr VI int line */
#endif
                        return(mem.null_mem);

                    case 0x14:   /* VI_BURST_REG */
                    case 0x15:
                    case 0x16:
                    case 0x17:
                        break;
                    
                    case 0x18:   /* VI_V_SYNC_REG */
                    case 0x19:
                    case 0x1a:
                    case 0x1b:
                        break;
                    
                    case 0x1c:   /* VI_H_SYNC_REG */
                    case 0x1d:
                    case 0x1e:
                    case 0x1f:
                        break;
                    
                    case 0x20:   /* VI_H_SYNC_LEAP_REG */
                    case 0x21:
                    case 0x22:
                    case 0x23:
                        break;
                    
                    case 0x24:   /* VI_H_START_REG */
                    case 0x25:
                    case 0x26:
                    case 0x27:
                        break;
                    
                    case 0x28:   /* VI_V_START_REG */
                    case 0x29:
                    case 0x2a:
                    case 0x2b:
                        break;
                    
                    case 0x2c:   /* VI_V_BURST_REG */
                    case 0x2d:
                    case 0x2e:
                    case 0x2f:
                        break;
                    
                    case 0x30:   /* VI_X_SCALE_REG */
                    case 0x31:
                    case 0x32:
                    case 0x33:
                        break;
                    
                    case 0x34:   /* VI_Y_SCALE_REG */
                    case 0x35:
                    case 0x36:
                    case 0x37:
                        break;
                    
                } /* switch(virtualaddr & 0xff) */

                return(mem.vi_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_8450_845f(uWORD virtualaddr)   /* ai_reg */
{
        static sWORD ai_addr_len_buffer_full = 0;

        if((virtualaddr & 0x1fffffff) < 0x04500018)
        {
                print_debug_msg(prefs.debug & DBG_AI_REG, "write - ai reg");

                switch(virtualaddr & 0xff)
                {
                    case 0x00:   /* AI_DRAM_ADDR_REG */
                    case 0x01:
                    case 0x02:
                    case 0x03:
                        if(ai_addr_len_buffer_full == 0)
                                ai_addr_len_buffer_full = 1;
                        else if(ai_addr_len_buffer_full == 1)
                                ai_addr_len_buffer_full = 1;
                        break;
                    
                    case 0x04:   /* AI_LEN_REG */
                    case 0x05:
                    case 0x06:
                    case 0x07:
                        if(ai_addr_len_buffer_full == 0)
                                ai_addr_len_buffer_full = 0;
                        else if(ai_addr_len_buffer_full == 1)
                                ai_addr_len_buffer_full = 2;
                        break;
                    
                    case 0x08:   /* AI_CONTROL_REG */
                    case 0x09:
                    case 0x0a:
                    case 0x0b:
                        if(ai_addr_len_buffer_full == 2)
                                ai_addr_len_buffer_full = 3;
                        mem_todo |= CHECK_AI_REG;
                        break;

                    case 0x0c:   /* AI_STATUS_REG */
                    case 0x0d:
                    case 0x0e:
                    case 0x0f:
#ifndef PUR03A_COMPATIBLE
                        ((uWORD *)mem.mi_reg)[2] &= ~MI_INTR_AI;   /* clr AI int line */
                        reg.do_or_check_sthg &= ~RS4300I_DO_AI_INTERRUPT;   /* from jeff */
#endif
                        return(mem.null_mem);

                    case 0x10:   /* AI_DACRATE_REG */
                    case 0x11:
                    case 0x12:
                    case 0x13:
                        break;

                    case 0x14:   /* AI_BITRATE_REG */
                    case 0x15:
                    case 0x16:
                    case 0x17:
                        break;
                        
                } /* switch(virtualaddr & 0xff) */

                return(mem.ai_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_8460_846f(uWORD virtualaddr)   /* pi_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04600034)
        {
                print_debug_msg(prefs.debug & DBG_PI_REG, "write - pi reg");
                
                switch(virtualaddr & 0xff)
                {
                    case 0x0c:   /* PI_WR_LEN_REG */
                    case 0x0d:
                    case 0x0e:
                    case 0x0f:
                        mem_todo |= PI_DMA;
                        break;
                        
                    case 0x10:   /* PI_STATUS_REG */
                    case 0x11:
                    case 0x12:
                    case 0x13:
                        mem_todo |= UPDATE_PI_REG;
                        return((char *)&(mem.pi_reg_status_w) + (virtualaddr & 0x3));
                        
                } /* switch(virtualaddr & 0xff) */

                return(mem.pi_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_8470_847f(uWORD virtualaddr)   /* ri_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x04700020)
        {
                print_debug_msg(prefs.debug & DBG_RI_REG, "write - ri reg");
                return(mem.ri_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_8480_848f(uWORD virtualaddr)   /* si_reg */
{
        if((virtualaddr & 0x1fffffff) < 0x0480001c)
        {
                print_debug_msg(prefs.debug & DBG_SI_REG, "write - si reg");

                switch(virtualaddr & 0xff)
                {
                    case 0x00:   /* SI_DRAM_ADDR_REG */
                    case 0x01:
                    case 0x02:
                    case 0x03:
                        break;
         
                    case 0x04:   /* SI_PIF_ADDR_RD64B_REG */
                    case 0x05:
                    case 0x06:
                    case 0x07:
                        mem_todo |= PIF2DRAM_DMA;
                        break;
         
                    case 0x08:   /* reserved */
                    case 0x09:
                    case 0x0a:
                    case 0x0b:
                        return(mem.null_mem);
         
                    case 0x0c:   /* reserved */
                    case 0x0d:
                    case 0x0e:
                    case 0x0f:
                        return(mem.null_mem);
         
                    case 0x10:   /* SI_PIF_ADDR_WR64B_REG */
                    case 0x11:
                    case 0x12:
                    case 0x13:
                        mem_todo |= DRAM2PIF_DMA;
                        break;
         
                    case 0x14:   /* reserved */
                    case 0x15:
                    case 0x16:
                    case 0x17:
                        return(mem.null_mem);
         
                    case 0x18:   /* SI_STATUS_REG */
                    case 0x19:
                    case 0x1a:
                    case 0x1b:
                        ((uWORD *)mem.mi_reg)[2] &= ~MI_INTR_SI;   /* clr SI intr in MI reg */
                        ((uWORD *)mem.si_reg)[6] &= ~0x00000001;   /* BUG:? */
                        return(mem.null_mem);

                } /* switch(virtualaddr & 0xff) */

                return(mem.si_reg + (virtualaddr & 0xff));
        }
        else
                return(get_electricalnoise(virtualaddr));
}

static char *put_mem_9fc0_9fcf(uWORD virtualaddr)   /* pif ram/rom */
{
        if((virtualaddr & 0x1fffffff) < 0x1fc007c0)
        {
                print_debug_msg(prefs.debug & DBG_PI_ROM, "write");
                return(mem.pi_rom + (virtualaddr & 0xfff));
        }
        else if((virtualaddr & 0x1fffffff) < 0x1fc00800)
        {
                print_debug_msg(prefs.debug & DBG_PI_RAM, "write");
                return(mem.pi_ram + (virtualaddr & 0x1fffffff) - 0x1fc007c0);
        }
        else
                return(get_electricalnoise(virtualaddr));
}










/*****************************************************************************\
*                                                                             *
* The static routine used by alloc_n64_mem().                                 *
*                                                                             *
\*****************************************************************************/

static void write_mem_addr()
{

        mem.rd_ram      = mem_addr[0];
        mem.sp_dmem     = mem_addr[1];
        mem.sp_imem     = mem_addr[2];
        mem.pi_rom      = mem_addr[3];
        mem.pi_ram      = mem_addr[4];
        mem.pi_ram_w    = mem_addr[5];
                     
        mem.rd_reg      = mem_addr[6];
        mem.sp_reg      = mem_addr[7];
        mem.dpc_reg     = mem_addr[8];
        mem.dps_reg     = mem_addr[9];
        mem.mi_reg      = mem_addr[10];
        mem.mi_reg_w    = mem_addr[11];
        mem.vi_reg      = mem_addr[12];
        mem.ai_reg      = mem_addr[13];
        mem.pi_reg      = mem_addr[14];
        mem.pi_reg_w    = mem_addr[15];
        mem.ri_reg      = mem_addr[16];
        mem.si_reg      = mem_addr[17];
        mem.tmem        = mem_addr[18];
        mem.null_mem    = mem_addr[19];

#ifdef SHM

        shminfo.rd_ram   = mem_shminfo[0];
        shminfo.sp_dmem  = mem_shminfo[1];
        shminfo.sp_imem  = mem_shminfo[2];
        shminfo.pi_rom   = mem_shminfo[3];
        shminfo.pi_ram   = mem_shminfo[4];
        shminfo.pi_ram_w = mem_shminfo[5];
                         
        shminfo.rd_reg   = mem_shminfo[6];
        shminfo.sp_reg   = mem_shminfo[7];
        shminfo.dpc_reg  = mem_shminfo[8];
        shminfo.dps_reg  = mem_shminfo[9];
        shminfo.mi_reg   = mem_shminfo[10];
        shminfo.mi_reg_w = mem_shminfo[11];
        shminfo.vi_reg   = mem_shminfo[12];
        shminfo.ai_reg   = mem_shminfo[13];
        shminfo.pi_reg   = mem_shminfo[14];
        shminfo.pi_reg_w = mem_shminfo[15];
        shminfo.ri_reg   = mem_shminfo[16];
        shminfo.si_reg   = mem_shminfo[17];
        shminfo.tmem     = mem_shminfo[18];
        shminfo.null_mem = mem_shminfo[19];

#endif /* SHM */

} /* static void write_mem_addr() */










/*****************************************************************************\
*                                                                             *
* The static routines used in final_check() - after a write to the n64 mem.   *
*                                                                             *
\*****************************************************************************/

static void do_display_refresh()
{
        dispatch(DISPLAY_REFRESH, 0);
        
} /* static void do_display_refresh(); */





#ifdef PUR_VI_INTERRUPT

static void check_vi_reg()
{
        /* BUG: the order might be changed. most changing reg first. (speed) */
        if( old_vi_status_reg != ((uWORD *)mem.vi_reg)[0] )
        {
                old_vi_status_reg = ((uWORD *)mem.vi_reg)[0];
#if DISPLAY_SUPPORT
                dispatch(DISPLAY_REFRESH, 0);
#endif
                return;
        }
        
        if( old_vi_width_reg  != ((uWORD *)mem.vi_reg)[2] )
        {
                old_vi_width_reg  = ((uWORD *)mem.vi_reg)[2];
#if DISPLAY_SUPPORT
                dispatch(DISPLAY_REFRESH, 0);
#endif
                return;
        }

} /* static void check_vi_reg() */

#endif /* #ifdef PUR_VI_INTERRUPT */




static void check_ai_reg()
{
        dispatch(AUDIO_UPDATE, 0);

} /* static void update_ai_reg() */





static void update_vi_reg()
{
        /* there are some unused bits in this regs.
           hope that this works */

        ((uWORD *)mem.vi_reg)[1] &= 0x003fffff;
        ((uWORD *)mem.vi_reg)[2] &= 0x00000fff;
        //printf("VI addr reg: %lx\n", ((uWORD *)mem.vi_reg)[1]);

} /* static void update_vi_reg() */





static void update_pi_reg()
{
        if(mem.pi_reg_status_w & 0x02)
                ((uWORD *)mem.mi_reg)[2] &= ~MI_INTR_PI;

} /* static void update_pi_reg() */





static void update_mi_reg()
{
        if(mem.mi_reg_mode_w & 0x0080)        /* MI_CLR_INIT */
                ((uWORD *)mem.mi_reg)[0] &= ~0x00000080;

        if(mem.mi_reg_mode_w & 0x0100)        /* MI_SET_INIT */
                ((uWORD *)mem.mi_reg)[0] |= 0x00000080;

        if(mem.mi_reg_mode_w & 0x0200)        /* MI_CLR_EBUS */
                ((uWORD *)mem.mi_reg)[0] &= ~0x00000100;

        if(mem.mi_reg_mode_w & 0x0400)        /* MI_SET_EBUS */
                ((uWORD *)mem.mi_reg)[0] |= 0x00000100;

        if(mem.mi_reg_mode_w & 0x0800)        /* MI_CLR_DP_INTR */
                ((uWORD *)mem.mi_reg)[2] &= ~MI_INTR_DP;

        if(mem.mi_reg_mode_w & 0x1000)        /* MI_CLR_RDRAM */
                ((uWORD *)mem.mi_reg)[0] &= ~0x00000200;

        if(mem.mi_reg_mode_w & 0x2000)        /* MI_SET_RDRAM */
                ((uWORD *)mem.mi_reg)[0] |= 0x00000200;



        if(mem.mi_reg_intr_mask_w & 0x0001)        /* MI_INTR_MASK_CLR_SP */
                ((uWORD *)mem.mi_reg)[3] &= ~0x0000001;

        if(mem.mi_reg_intr_mask_w & 0x0002)        /* MI_INTR_MASK_SET_SP */
                ((uWORD *)mem.mi_reg)[3] |= 0x00000001;

        if(mem.mi_reg_intr_mask_w & 0x0004)        /* MI_INTR_MASK_CLR_SI */
                ((uWORD *)mem.mi_reg)[3] &= ~0x0000002;

        if(mem.mi_reg_intr_mask_w & 0x0008)        /* MI_INTR_MASK_SET_SI */
                ((uWORD *)mem.mi_reg)[3] |= 0x00000002;

        if(mem.mi_reg_intr_mask_w & 0x0010)        /* MI_INTR_MASK_CLR_AI */
                ((uWORD *)mem.mi_reg)[3] &= ~0x00000004;

        if(mem.mi_reg_intr_mask_w & 0x0020)        /* MI_INTR_MASK_SET_AI */
                ((uWORD *)mem.mi_reg)[3] |= 0x00000004;

        if(mem.mi_reg_intr_mask_w & 0x0040)        /* MI_INTR_MASK_CLR_VI */
                ((uWORD *)mem.mi_reg)[3] &= ~0x00000008;

        if(mem.mi_reg_intr_mask_w & 0x0080)        /* MI_INTR_MASK_SET_VI */
                ((uWORD *)mem.mi_reg)[3] |= 0x00000008;

        if(mem.mi_reg_intr_mask_w & 0x0100)        /* MI_INTR_MASK_CLR_PI */
                ((uWORD *)mem.mi_reg)[3] &= ~0x00000010;

        if(mem.mi_reg_intr_mask_w & 0x0200)        /* MI_INTR_MASK_SET_PI */
                ((uWORD *)mem.mi_reg)[3] |= 0x00000010;

        if(mem.mi_reg_intr_mask_w & 0x0400)        /* MI_INTR_MASK_CLR_DP */
                ((uWORD *)mem.mi_reg)[3] &= ~0x00000020;

        if(mem.mi_reg_intr_mask_w & 0x0800)        /* MI_INTR_MASK_SET_DP */
                ((uWORD *)mem.mi_reg)[3] |= 0x00000020;

} /* static void update_mi_reg() */






static void update_sp_reg()
{
        if(mem.sp_reg_status_w & 0x00000001)   /* SP_CLR_HALT       */
        {
                ((uWORD *)mem.sp_reg)[4] &= ~0x00000001;
                rsp_reg.halt = 0;
        }

        if(mem.sp_reg_status_w & 0x00000002)   /* SP_SET_HALT       */
        {
                ((uWORD *)mem.sp_reg)[4] |= 0x00000001;
                rsp_reg.halt = 1;
        }

        if(mem.sp_reg_status_w & 0x00000004)   /* SP_CLR_BROKE      */
                ((uWORD *)mem.sp_reg)[4] &= ~0x00000002;

        if(mem.sp_reg_status_w & 0x00000008)   /* SP_CLR_INTR       */
                ((uWORD *)mem.mi_reg)[2] &= ~MI_INTR_SP;

        if(mem.sp_reg_status_w & 0x00000010)   /* SP_SET_INTR       */
        {
                ((uWORD *)mem.mi_reg)[2] |= MI_INTR_SP;
                reg.do_or_check_sthg = RS4300I_CHECK_MI_INTERRUPTS;
        }

        if(mem.sp_reg_status_w & 0x00000020)   /* SP_CLR_SSTEP      */
        {
                ((uWORD *)mem.sp_reg)[4] &= ~0x00000020;
        }

        if(mem.sp_reg_status_w & 0x00000040)   /* SP_SET_SSTEP      */
        {
                ((uWORD *)mem.sp_reg)[4] |= 0x00000020;
        }

        if(mem.sp_reg_status_w & 0x00000080)   /* SP_CLR_INTR_BREAK */
                ((uWORD *)mem.sp_reg)[4] &= ~0x00000040;

        if(mem.sp_reg_status_w & 0x00000100)   /* SP_SET_INTR_BREAK */
                ((uWORD *)mem.sp_reg)[4] |= 0x00000040;

        if(mem.sp_reg_status_w & 0x00000200)   /* SP_CLR_SIG0       */
                ((uWORD *)mem.sp_reg)[4] &= ~0x00000080;

        if(mem.sp_reg_status_w & 0x00000400)   /* SP_SET_SIG0       */
                ((uWORD *)mem.sp_reg)[4] |= 0x00000080;

        if(mem.sp_reg_status_w & 0x00000800)   /* SP_CLR_SIG1       */
                ((uWORD *)mem.sp_reg)[4] &= ~0x00000100;

        if(mem.sp_reg_status_w & 0x00001000)   /* SP_SET_SIG1       */
                ((uWORD *)mem.sp_reg)[4] |= 0x00000100;

        if(mem.sp_reg_status_w & 0x00002000)   /* SP_CLR_SIG2       */
                ((uWORD *)mem.sp_reg)[4] &= ~0x00000200;

        if(mem.sp_reg_status_w & 0x00004000)   /* SP_SET_SIG2       */
                ((uWORD *)mem.sp_reg)[4] |= 0x00000200;

        if(mem.sp_reg_status_w & 0x00008000)   /* SP_CLR_SIG3       */
                ((uWORD *)mem.sp_reg)[4] &= ~0x00000400;

        if(mem.sp_reg_status_w & 0x00010000)   /* SP_SET_SIG3       */
                ((uWORD *)mem.sp_reg)[4] |= 0x00000400;

        if(mem.sp_reg_status_w & 0x00020000)   /* SP_CLR_SIG4       */
                ((uWORD *)mem.sp_reg)[4] &= ~0x00000800;

        if(mem.sp_reg_status_w & 0x00040000)   /* SP_SET_SIG4       */
                ((uWORD *)mem.sp_reg)[4] |= 0x00000800;

        if(mem.sp_reg_status_w & 0x00080000)   /* SP_CLR_SIG5       */
                ((uWORD *)mem.sp_reg)[4] &= ~0x00001000;

        if(mem.sp_reg_status_w & 0x00100000)   /* SP_SET_SIG5       */
                ((uWORD *)mem.sp_reg)[4] |= 0x00001000;

        if(mem.sp_reg_status_w & 0x00200000)   /* SP_CLR_SIG6       */
                ((uWORD *)mem.sp_reg)[4] &= ~0x00002000;

        if(mem.sp_reg_status_w & 0x00400000)   /* SP_SET_SIG6       */
                ((uWORD *)mem.sp_reg)[4] |= 0x00002000;

        if(mem.sp_reg_status_w & 0x00800000)   /* SP_CLR_SIG7       */
                ((uWORD *)mem.sp_reg)[4] &= ~0x00004000;

        if(mem.sp_reg_status_w & 0x01000000)   /* SP_SET_SIG7       */
                ((uWORD *)mem.sp_reg)[4] |= 0x00004000;

} /* static void update_sp_reg() */





static void update_dp_reg()
{
        if(mem.dpc_reg_status_w & 0x0001)       /* DPC_CLR_XBUS_DMEM_DMA */
                ((uWORD *)mem.dpc_reg)[3] &= ~0x0000001;

        if(mem.dpc_reg_status_w & 0x0002)       /* DPC_SET_XBUS_DMEM_DMA */
                ((uWORD *)mem.dpc_reg)[3] |= 0x00000001;

        if(mem.dpc_reg_status_w & 0x0004)       /* DPC_CLR_FREEZE */
                ((uWORD *)mem.dpc_reg)[3] &= ~0x0000002;

        if(mem.dpc_reg_status_w & 0x0008)       /* DPC_SET_FREEZE */
                ((uWORD *)mem.dpc_reg)[3] |= 0x00000002;

        if(mem.dpc_reg_status_w & 0x0010)       /* DPC_CLR_FLUSH */
                ((uWORD *)mem.dpc_reg)[3] &= ~0x0000004;

        if(mem.dpc_reg_status_w & 0x0020)       /* DPC_SET_FLUSH */
                ((uWORD *)mem.dpc_reg)[3] |= 0x00000004;

        if(mem.dpc_reg_status_w & 0x0040)       /* DPC_CLR_TMEM_CTR */
                ((uWORD *)mem.dpc_reg)[7] = 0;

        if(mem.dpc_reg_status_w & 0x0080)       /* DPC_CLR_PIPE_CTR */
                ((uWORD *)mem.dpc_reg)[6] = 0;

        if(mem.dpc_reg_status_w & 0x0100)       /* DPC_CLR_CMD_CTR */
                ((uWORD *)mem.dpc_reg)[5] = 0;

        if(mem.dpc_reg_status_w & 0x0200)       /* DPC_CLR_CLOCK_CTR */
                ((uWORD *)mem.dpc_reg)[4] = 0;

} /* static void update_dp_reg() */






