/**
 * TrueReality - output/display/x11/x11.c
 * Copyright (C) 1998, 1999 by Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <signal.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#ifdef OPENGL
        #include <GL/gl.h>
        #include <GL/glut.h>
        #include <GL/glx.h>     /* X11 specific */
#endif

#include <config.h>
#include <parser_extern.h>
#include <type_sizes.h>
#include <memory_extern.h>
#include <dispatch.h>
#include <x11_extern.h>

#include "../display.h"
#include "../warp_resolution_and_depth.h"
#include "x11.h"
#include "../../../N64/registers.h"
#include "../../../N64/rdp_registers.h"

#ifdef DEBUG
        #include <debug.h>
#endif

#ifdef SHM
        #include <sys/ipc.h>
        #include <sys/shm.h>
        #include <X11/extensions/XShm.h>
#endif

#ifdef FPS
#ifdef LINUX   /* FPS - I don't know if this works on other os than linux - please mail */
        #include <sys/time.h>
        #include <unistd.h>
#endif
#endif





/* the following 2 variables are exported in 'include/x11_extern.h' */
Display                         *dsp;
Window                          win;
int                             valid_clip_rect = 0;
#ifndef JUST3D
        static Pixmap           back_pix;
        static GC               back_gc;
        static Pixmap           tile_pix[8];
#endif
static XRectangle               back_cliprect[1];

static Screen                   *scr;
static XWindowAttributes        win_attrib;
static XSetWindowAttributes     win_setattrib;
XImage                          *img;
XImage                          *tile_img[8];
int                             tile_img_used[8] = {0, 0, 0, 0, 0, 0, 0, 0};
static GC                       gc;
static unsigned long            white, black;
static XColor                   color;
XColor                          color6bit[64];
//static XColor                 color8bit[256];   /* by Niki W. Waibel */
static Colormap                 colormap;

static char                     title[] = "TrueReality-" VERSION " by Niki W. Waibel and the help of many others";
static uWORD                     width  =   0;   /* initial VI_WIDTH_REG (0x4400008) value */
static uWORD                     height =   0;   /* width*3/4 */
static int                      WIDTH  = 640;   /* real width of window */
static int                      HEIGHT = 480;   /* real height of window */
#ifdef JUST3D
        static float            stretch_x;      /* x stretch factor (WIDTH/width) */
        static float            stretch_y;      /* y stretch factor (HEIGHT/width*3/4) */
#endif
#undef DEPTH
static int                      DEPTH;
static int                      REAL_DEPTH;

static void                     *XBuf  = NULL;
static uWORD                     flags  = 0;

#ifdef SHM
        static XShmSegmentInfo  XBuf_SHMInfo;
#endif

#ifndef JUST3D
        static int  rdp_valid_addr;
#endif
        static char *rdp_address;
        static int  rdp_format;
        static int  rdp_size;
        static int  rdp_width;







static  int     InitDisplay();
static  int     DestroyDisplay();

static  void    PrintText(char *text, int x, int y);

#ifdef DEBUG
        static  void PrintDisplayMode(uWORD flags);
#endif

static  void    PutImage(int x, int y, int w, int h);
#ifndef JUST3D
static  void    PutBackPixmap();
#endif
#ifndef JUST3D
#ifdef OPENGL
void            DrawScreenGL(int width, int height);
#endif
#endif

#ifdef WE_DONT_NEED_THIS
static  int     SetActiveTexture();
#endif

static  void    AllocDisplayMem();
static  void    FreeDisplayMem();





#ifdef OPENGL

static int attrib[] =
{
        GLX_RGBA,
        GLX_RED_SIZE, 1,
        GLX_BLUE_SIZE, 1,
        GLX_GREEN_SIZE, 1,
        GLX_DEPTH_SIZE, 2,
#ifdef JUST3D
        GLX_DOUBLEBUFFER,
#endif
        None
};

static XVisualInfo      *vis;
static GLXContext       ctx;

#ifndef JUST3D
        static Pixmap   gl_pix;
#endif

#endif /* #ifdef OPENGL */





/* choose renderer (directx is not used - but that way you can implement it in windows) */

#ifndef OPENGL
#ifndef DIRECTX
#include "../sw_renderer.h"
#endif
#endif

#ifdef OPENGL
#include "../gl_renderer.h"   /* OpenGL stuff */
#endif

#ifdef DIRECTX
#include "../dx_renderer.h"   /* DirectX stuff */
#endif





/*******************************************************************************
*                                                                              *
* int OpenVisual()                                                             *
* int CloseVisual()                                                            *
*                                                                              *
********************************************************************************
*                                                                              *
* These routines are used by the emulator to open and to close a visual.       *
*                                                                              *
* Both return a negative int on error and zero on success.                     *
*                                                                              *
*******************************************************************************/





int OpenVisual()
{
        if(InitDisplay() < 0)
                return(-1);

    /* get the global keyboard mode (autorepeat on or off) */
        XGetKeyboardControl(dsp, &xkeyboardstate);

        prefs.display = LOCAL_DISPLAY;

#ifdef OPENGL
        CreateDisplayLists();
#endif
        return(0);

} /* int OpenVisual() */





int CloseVisual()
{
        DestroyDisplay();

        prefs.display = NO_DISPLAY;

        return(0);

} /* int CloseVisual() */










/*******************************************************************************
*                                                                              *
* static int InitDisplay()                                                     *
*                                                                              *
********************************************************************************
*                                                                              *
* This routine is used to initialize the visual.                               *
* It should print serious information of the current status.                   *
* It must NOT be called by the emulator. It's called from 'int OpenVisual()'.  *
* The name is not standard but it should be used!                              *
*                                                                              *
* The routine returns a negative int on error and zero on success.             *
*                                                                              *
*******************************************************************************/





static int InitDisplay()
{
        XEvent                  event;
        int                     i;
        unsigned long           mask;

#ifdef SHM
        struct shmid_ds         s;
#endif




        puts("Initializing Unix/X display: (most stuff by \"Guenther Sohler\")");



        fputs("Opening display ...", stdout); fflush(stdout);

        dsp = XOpenDisplay(NULL);
        if(!dsp)
        {
                puts(" FAILED\n"); fflush(stdout);
                return(-1);
        }

        scr     = DefaultScreenOfDisplay(dsp);

        white   = WhitePixelOfScreen(scr);
        black   = BlackPixelOfScreen(scr);

#ifdef OPENGL
        fputs(" OK\n"
              "Attaching OpenGL (Mesa/glX) extensions ...", stdout); fflush(stdout);
        vis = glXChooseVisual(dsp, DefaultScreen(dsp), attrib);
        if(!vis)
        {
                puts(" FAILED\n"); fflush(stdout);
                return(-1);
        }
#endif

        fputs(" OK\n"
              "Creating window ...", stdout); fflush(stdout);

        win_setattrib.backing_store = WhenMapped;
        win_setattrib.background_pixel = 0;
        win_setattrib.border_pixel = 0;
        win_setattrib.event_mask = StructureNotifyMask | ExposureMask;
        mask = CWBackingStore | CWBackPixel | CWBorderPixel | CWEventMask;

#ifdef OPENGL
        win = XCreateWindow(dsp, RootWindowOfScreen(scr),
                0, 0, WIDTH, HEIGHT,
                0, vis->depth, InputOutput, vis->visual, mask, &win_setattrib
                );
#else
        win = XCreateWindow(dsp, RootWindowOfScreen(scr),
                0, 0, WIDTH, HEIGHT,
                0, CopyFromParent, InputOutput, CopyFromParent, mask, &win_setattrib
                );
#endif
        if(!win)
        {               
                puts(" FAILED\n"); fflush(stdout);
                return(-1);
        }


/* all of the DEPTH stuff is from "Guenther Sohler" <guenther.sohler@newlogic.at> */

        fputs(" OK\n"
              "Checking depth of window ...", stdout); fflush(stdout);

/* Niki W. Waibel: the following does not differ between 24bit and 32bit! */
/* ... but it's ok for 'XCreateImage'                                     */
/* -> we can get the real depth after 'XCreateImage'                      */

        XGetWindowAttributes(dsp, win, &win_attrib);
        DEPTH = win_attrib.depth;

        printf(" %02dbit ... OK\n", DEPTH); fflush(stdout);

        if(DEPTH == 8)
        {
                fputs("With 8 bit we've to allocate 64 colors ...", stdout); fflush(stdout);
                colormap = XCreateColormap(dsp, win, win_attrib.visual, AllocNone);
                XSetWindowColormap(dsp, win, colormap);

                for(i=0; i<64; i++)
                {
                        color.blue  = (i & 0x03) << 14;
                        color.green = (i & 0x0C) << 12;
                        color.red   = (i & 0x30) << 10;
                        if( !XAllocColor(dsp, colormap, &color) )
                        {
                                puts(" FAILED\n"); fflush(stdout);
                                return(-1);
                        }
                        color6bit[i] = color;

                } /* for(i=0; i<64; i++) */

                puts(" OK"); fflush(stdout);

        } /* if(DEPTH == 8) */



        {
                XSizeHints      Hints;
                XWMHints        WMHints;


                Hints.flags       = PSize | PMinSize | PMaxSize;
                Hints.min_width   = 32;
                Hints.max_width   = 1280;
                Hints.base_width  = WIDTH;
                Hints.min_height  = 24;
                Hints.max_height  = 1024;
                Hints.base_height = HEIGHT;
                WMHints.input     = True;
                WMHints.flags     = InputHint;

                XSetWMHints(dsp, win, &WMHints);
                XSetWMNormalHints(dsp, win, &Hints);
                XStoreName(dsp, win, title);
        }




        /* input events */

        XSelectInput(dsp, win, FocusChangeMask | ExposureMask);



        fputs("Mapping window ...", stdout); fflush(stdout);

        XMapRaised(dsp, win);

        fputs(" OK\n"
              "Clearing window ...", stdout); fflush(stdout);

        XClearWindow(dsp, win);

        /* catch first event to display window */
        XWindowEvent(dsp, win, ExposureMask, &event);



        fputs(" OK\n"
              "Creating graphics context ...", stdout); fflush(stdout);

        gc      = XCreateGC(dsp, win, 0, 0);




#ifdef SHM   /* this code was first coded by me (Niki W. Waibel) */
             /* then improved by Guenther Sohler (thanks a lot, Gue!) */
        if(prefs.UseSHM)
        {
                fputs(" OK\n"
                      "--- SHM stuff ---\n"
                      "Creating image in shared memory ...", stdout); fflush(stdout);
      
                img = XShmCreateImage(dsp, DefaultVisualOfScreen(scr), DEPTH, ZPixmap, NULL, &XBuf_SHMInfo, WIDTH, HEIGHT);
                if(!img)
                { 
                        puts(" FAILED\n"); fflush(stdout);
                        return(-1);
                }
      
                fputs(" OK\n"
                      "Allocating display memory buffer as shared memory ...", stdout);
                fflush(stdout);
      
                XBuf_SHMInfo.shmid = shmget(IPC_PRIVATE, img->height * img->bytes_per_line, IPC_CREAT | 0777);
      
                if(XBuf_SHMInfo.shmid < 0)
                {
                        puts(" FAILED\n"); fflush(stdout);
                        return(-1);
                }
      
                fputs(" OK\n"
                      "Attaching shared memory (shmat) ...", stdout); fflush(stdout);

                XBuf = img->data = XBuf_SHMInfo.shmaddr = shmat(XBuf_SHMInfo.shmid, 0, 0);
                if(!XBuf)
                {
                        puts(" FAILED (shmat)\n"); fflush(stdout);
                        return(-1);
                }

                XBuf_SHMInfo.readOnly = False;
      
                fputs(" OK\n"                     
                      "Attaching shared memory (XShmAttach) ...", stdout); fflush(stdout);
      
                if(!XShmAttach(dsp, &XBuf_SHMInfo))
                {
                        puts(" FAILED\n"); fflush(stdout);
                        return(-1); 
                }         

                shmctl(XBuf_SHMInfo.shmid, IPC_RMID, &s);   /* if someone kills me shm should be freed */
             
                puts(" OK\n"
                     "--- end of SHM stuff ---"); fflush(stdout);
        }
        else
#endif
        {
                fputs(" OK\n"
                      "Creating image ...", stdout); fflush(stdout);

                img = XCreateImage(dsp, DefaultVisualOfScreen(scr), DEPTH, ZPixmap, 0, XBuf, WIDTH, HEIGHT, DEPTH, 0);
                if(!img)
                { 
                        puts(" FAILED\n"); fflush(stdout);
                        return(-1);
                }

                fputs(" OK\n"
                      "Allocating display memory buffer ...", stdout); fflush(stdout);

                XBuf = img->data = malloc(img->height * img->bytes_per_line);
                if(!XBuf)
                {
                        puts(" FAILED\n"); fflush(stdout);
                        return(-1);
                }
                puts(" OK"); fflush(stdout);

        }

        fputs("Checking depth of image ...", stdout); fflush(stdout);
        REAL_DEPTH = img->bits_per_pixel;
        printf(" %02dbit ... OK\n", REAL_DEPTH); fflush(stdout);

        if(DEPTH == 15)
                REAL_DEPTH = 15;


#ifndef JUST3D
        fputs("Creating 3D doublebuffer render pixmap ...", stdout); fflush(stdout);

        rdp_valid_addr = 0;

#ifndef OPENGL
        back_pix = XCreatePixmap(dsp, RootWindowOfScreen(scr), WIDTH, HEIGHT, DEPTH);
        back_gc  = XCreateGC(dsp, back_pix, 0, 0);
#endif

        puts(" OK"); fflush(stdout);
#endif



#ifdef OPENGL
        ctx = glXCreateContext(dsp, vis, NULL, True);
#ifdef JUST3D
        glXMakeCurrent(dsp, win, ctx);
#else
        gl_pix = glXCreateGLXPixmapMESA(dsp, vis, back_pix, win_attrib.colormap);
        glXMakeCurrent(dsp, gl_pix, ctx);
#endif
#endif




        printf("-------------------------------------------------------------------------------\n"); fflush(stdout);

        RefreshVisual();

        return(0);

} /* static int InitDisplay() */










/*******************************************************************************
*                                                                              *
* static int DestroyDisplay()                                                  *
*                                                                              *
********************************************************************************
*                                                                              *
* This routine is used to destroy the visual.                                  *
* It should print serious information of the current status.                   *
* It must NOT be called by the emulator. It's called from 'int CloseVisual()'. *
* The name is not standard but it should be used!                              *
*                                                                              *
* The routine returns a negative int on error and zero on success.             *
*                                                                              *
*******************************************************************************/





static int DestroyDisplay()
{
        printf("-------------------------------------------------------------------------------\n");

        printf("Destroying Unix/X display:\n");

#ifdef SHM
        if(prefs.UseSHM)
        {
                printf("Detaching shared memory ..."); fflush(stdout);

                XShmDetach(dsp, &XBuf_SHMInfo);
                shmdt(XBuf_SHMInfo.shmaddr);
                free(img);
        }
        else
#endif /* #ifdef SHM */
        {
                printf("Destroying image ..."); fflush(stdout);
                
                XDestroyImage(img);
        }

        printf(" OK\n"
               "Freeing graphics context ..."); fflush(stdout);

        XFreeGC(dsp, gc);

#ifdef OPENGL
        printf(" OK\n"
          "Destroying OpenGL (Mesa) stuff ..."); fflush(stdout);

        glXDestroyContext(dsp, ctx);

#ifndef JUST3D
        glXDestroyGLXPixmap(dsp, gl_pix);
#endif
#endif

#ifndef JUST3D
        rdp_valid_addr = 0;
#endif

#ifndef JUST3D
        printf(" OK\n"
               "Freeing 3D render pixmap ..."); fflush(stdout);
        XFreePixmap(dsp, back_pix);

        printf(" OK\n"
               "Freeing 3D graphics context ..."); fflush(stdout);
        XFreeGC(dsp, back_gc);
#endif

        printf(" OK\n"
               "Unmapping window ..."); fflush(stdout);

        XUnmapWindow(dsp, win);

        printf(" OK\n"
          "Destroying window ..."); fflush(stdout);

        XDestroyWindow(dsp, win);

        printf(" OK\n"
          "Closing display ..."); fflush(stdout);

        if(xkeyboardstate.global_auto_repeat == AutoRepeatModeOn)
                XAutoRepeatOn(dsp);

        XCloseDisplay(dsp);

        printf(" OK\n");



        printf("-------------------------------------------------------------------------------\n"); fflush(stdout);

        return(0);

} /* static int DestroyDisplay() */









static void PrintText(char *text, int x, int y)
{
        XSetForeground(dsp, gc, black);
        XDrawString(dsp, win, gc, x+1, y+1, text, strlen(text));
        
        XSetForeground(dsp, gc, white);
        XDrawString(dsp, win, gc, x, y, text, strlen(text));
   
} /* static void PrintText(char *text, int x, int y) */





#ifndef JUST3D
static void PrintBackText(char *text, int x, int y)
{
        XSetForeground(dsp, back_gc, black);
        XDrawString(dsp, back_pix, back_gc, x+1, y+1, text, strlen(text));
        
        XSetForeground(dsp, back_gc, white);
        XDrawString(dsp, back_pix, back_gc, x, y, text, strlen(text));
   
} /* static void PrintText(char *text, int x, int y) */
#endif




#ifdef DEBUG

static void PrintDisplayMode(uWORD flags)
{
        printf("Displaymode:\n");
   

   
        switch(flags & 0x0003)
        {
            case 0:
                printf("   blank (no data, no sync)\n");
                break;

            case 1:
                printf("   reserved\n");
                break;

            case 2:
                printf("   5/5/5/3 (\"16\" bit)\n");
                break;
         
            case 3:
                printf("   8/8/8/8 (32 bit)\n");
                break;
         
        } /* switch(flags & 0x0003) */
   

   
        if(flags & 0x0004)
                printf("   gamma_dither_enable\n");
        if(flags & 0x0008)
                printf("   gamma_enable\n");
        if(flags & 0x0010)
                printf("   divot_enable\n");
        if(flags & 0x0040)
                printf("   serrate\n");
   

   
        printf("   anti-alias (aa) mode\n");
   
        switch((flags & 0x0300) >> 8)
        {
            case 0:
                printf("      aa & resamp (always fetch extra lines)\n");
                break;

            case 1:
                printf("      aa & resamp (fetch extra lines if needed)\n");
                break;

            case 2:
                printf("      resamp only (treat as all fully covered)\n");
                break;
         
            case 3:
                printf("      neither (replicate pixels, no interpolate)\n");
                break;
         
        } /* switch((flags & 0x0300) >> 8) */
   
} /* static void PrintDisplayMode(uWORD flags) */

#endif





void RefreshVisual()
{
        uWORD    new_width;
        uBYTE    *imsrc1, *imdst1;
        uHWORD   *imsrc2, *imdst2;
        uWORD    *imsrc4, *imdst4;



#ifdef FPS
#ifdef LINUX   /* FPS - I don't know if this works on other os than linux - please mail */
        static char             fps_text[40];
        int                     fps_update_sec = 1;   /* min update frequency in sec */
                                                      /* otherwise you might not read it properly */
        static long             fps_update = 0;   /* stores the next time for update */
        static int              fps_count = 0;   /* to make fps average */
        static float            fps_rate = 0.0;
        static struct timeval   fps_tv_old  = { 0, 0 };
        struct timeval          fps_tv;
        
#endif
#endif



        /* video flags */

        flags = ((uWORD *)mem.vi_reg)[0];

#ifdef DEBUG
        if(prefs.debug & DBG_DISP_REFRESH)
                puts("DBG_DISP_REFRESH: refresh");
        if(prefs.debug & DBG_DISP_MODE)
                PrintDisplayMode(flags);
#endif



        /* video width */
        /* if width in VI_WIDTH_REG (0x04400008) has changed set windows width */

        new_width = ((uWORD *)mem.vi_reg)[2];
        if(new_width != width)
        {
                if(new_width < 320 || new_width > 640)
                        width = 640;
                else
                        width = new_width;

                height = (width * 3) >> 2;

                XResizeWindow(dsp, win, width, height);
                XFlush(dsp);
        }



        /* display it! */

#ifdef OPENGL
        DrawScreenGL(width, height);
        Visual3D_FPS();
        FlushVisualRenderBuffer();
        return;
#endif
        if(new_width >= 320 && new_width <= 640)
        {
                if(flags & 0x0003)
                {
                        XGetWindowAttributes(dsp, win, &win_attrib);
                        if(WIDTH != win_attrib.width || HEIGHT != win_attrib.height)
                        {
                                WIDTH  = win_attrib.width & 0xfffc;
                                HEIGHT = win_attrib.height;

                                /* free all display related memory */
                                FreeDisplayMem();

                                /* allocate a new portion of display related memory */
                                AllocDisplayMem();

                        } /* if(WIDTH != win_attrib.width || HEIGHT != win_attib.height) */



                        /* source video origin */
                        imsrc4 = (uWORD  *) (mem.rd_ram + ((uWORD *)mem.vi_reg)[1]);
                        imsrc2 = (uHWORD *) imsrc4;
                        imsrc1 = (uBYTE  *) imsrc2;

                        /* destination video origin */
                        imdst4 = (uWORD  *) img->data;
                        imdst2 = (uHWORD *) imdst4;
                        imdst1 = (uBYTE  *) imdst2;

                        switch(REAL_DEPTH)
                        {
                            case 8:
                                switch(flags & 0x0003)
                                {
                                    case 0:
                                    case 1:
                                        break;

                                    case 2:   /* local: 2/2/2 (8bit - 6bit used)  n64: 5/5/5/3 (16bit) */
                                        render_image_16_to_06(imsrc2, imdst1, width, height, WIDTH, HEIGHT);
                                        break;

                                    case 3:   /* local: 2/2/2 (8bit - 6bit used)   n64: 8/8/8/8 (32bit) */
                                        render_image_32_to_06(imsrc4, imdst1, width, height, WIDTH, HEIGHT);
                                        break;

                                } /* switch(flags & 0x0003) */
                                break;

                            case 16:
                                switch(flags & 0x0003)
                                {
                                    case 0:
                                    case 1:
                                        break;

                                    case 2:   /* local: 5/6/5 (16bit)   n64: 5/5/5/3 (16bit) */
                                        render_image_16_to_16(imsrc2, imdst2, width, height, WIDTH, HEIGHT);
                                        break;

                                    case 3:   /* local: 5/6/5 (16bit)   n64: 8/8/8/8 (32bit) */
                                        render_image_32_to_16(imsrc4, imdst2, width, height, WIDTH, HEIGHT);
                                        break;

                                } /* switch(flags & 0x0003) */
                                break;

                            case 15:
                                switch(flags & 0x0003)
                                {
                                    case 0:
                                    case 1:
                                        break;

                                    case 2:   /* local: 5/5/5 (15bit)   n64: 5/5/5/3 (16bit) */
                                        render_image_16_to_15(imsrc2, imdst2, width, height, WIDTH, HEIGHT);
                                        break;

                                    case 3:   /* local: 5/5/5 (15bit)   n64: 8/8/8/8 (32bit) */
                                        render_image_32_to_15(imsrc4, imdst2, width, height, WIDTH, HEIGHT);
                                        break;

                                } /* switch(flags & 0x0003) */
                                break;

                            case 24:
                                switch(flags & 0x0003)
                                {
                                    case 0:
                                    case 1:
                                        break;

                                    case 2:   /* local: 8/8/8 (24bit)   n64: 5/5/5/3 (16bit) */
                                        render_image_16_to_24(imsrc2, imdst1, width, height, WIDTH, HEIGHT);
                                        break;

                                    case 3:   /* local: 8/8/8 (24bit)   n64: 8/8/8/8 (32bit) */
                                        render_image_32_to_24(imsrc4, imdst1, width, height, WIDTH, HEIGHT);
                                        break;

                                } /* switch(flags & 0x0003) */
                                break;

                            case 32:
                                switch(flags & 0x0003)
                                {
                                    case 0:
                                    case 1:
                                        break;

                                    case 2:   /* local: 8/8/8 (32bit)   n64: 5/5/5/3 (16bit) */
                                        render_image_16_to_32(imsrc2, imdst4, width, height, WIDTH, HEIGHT);
                                        break;

                                    case 3:   /* local: 8/8/8 (32bit)   n64: 8/8/8/8 (32bit) */
                                        render_image_32_to_32(imsrc4, imdst4, width, height, WIDTH, HEIGHT);
                                        break;

                                } /* switch(flags & 0x0003) */
                                break;

                            default:
                                fprintf(stderr, "Display depth of %d bit is not supported! (yet ;) )\n", REAL_DEPTH);

                        } /* switch(DEPTH) */

                        PutImage(0, 0, WIDTH-1, HEIGHT-1);
                }
                else
                {
                        PrintText("DISPLAY: blank (no data, no sync)", 20, 40);
                }
        }
        else
        {
                if(!(flags & 0x0003))
                        PrintText("DISPLAY: blank (no data, no sync)", 20, 40);

                PrintText("DISPLAY: wrong width (640 will be used!)", 20, 60);
        }

#ifdef FPS
#ifdef LINUX   /* FPS - I don't know if this works on other os than linux - please mail */

        gettimeofday(&fps_tv, NULL);

        fps_rate = fps_count * fps_rate
                   + 1 / ( ((float)fps_tv.tv_sec     + (float)fps_tv.tv_usec     / 1000000)
                         - ((float)fps_tv_old.tv_sec + (float)fps_tv_old.tv_usec / 1000000) );

        fps_count++;

        fps_rate /= fps_count;

        if(fps_update <= fps_tv.tv_sec)
        {
                fps_update = fps_tv.tv_sec + fps_update_sec;
                fps_count  = 0;

                if(fps_rate > 99.9)
                        fps_rate = 99.99;

                sprintf(fps_text, "%05.2f", fps_rate);

                fps_rate = 0.0;
        }

        fps_tv_old.tv_sec = fps_tv.tv_sec;
        fps_tv_old.tv_usec = fps_tv.tv_usec;

        PrintText(fps_text, 01, 20);
#endif
#endif

        XFlush(dsp);

} /* void RefreshVisual() */





void GetVisualName(char *name, int *n)
{
        strncpy(name, DISPLAYNAME, (size_t) *n);

        *n = strlen(DISPLAYNAME);

} /* void GetVisualName() */





void SetVisualTitle(char *title)
{
        XStoreName(dsp, win, title);

} /* void SetVisualTitle() */







































static void PutImage(int X, int Y, int W, int H)
{
        //img->bitmap_bit_order = LSBFirst;
        //img->bitmap_bit_order = MSBFirst;

#ifdef SHM
        if(prefs.UseSHM)
        {
                XShmPutImage(dsp, win, gc, img, X, Y, (WIDTH-W)/2, (HEIGHT-H)/2, W, H, False);
        }
        else
#endif
                XPutImage(dsp, win, gc, img, X, Y, (WIDTH-W)/2, (HEIGHT-H)/2, W, H);

} /* static void PutImage(int X, int Y, int W, int H) */





static void FreeDisplayMem()
{
#ifdef SHM
        if(prefs.UseSHM)
        {
                XShmDetach(dsp, &XBuf_SHMInfo);
                shmdt(XBuf_SHMInfo.shmaddr);
                free(img);
        }
        else
#endif
        {
                XDestroyImage(img);
        }
        //fputs("FreeDisplayMem: ready\n", stderr); fflush(stderr);

} /* static void FreeDisplayMem() */





static void AllocDisplayMem()
{
#ifdef SHM
        struct shmid_ds s;
#endif



#ifdef SHM   /* this code was first coded by me (Niki W. Waibel) */
             /* then improved by Guenther Sohler (thanks a lot, Gue!) */
        if(prefs.UseSHM)
        {
                img = XShmCreateImage(dsp, DefaultVisualOfScreen(scr), DEPTH, ZPixmap, NULL, &XBuf_SHMInfo, WIDTH, HEIGHT);
                if(!img)
                { 
                        puts("FATAL DISPLAY ERROR: XCreateImage\n\n"); fflush(stdout);
                        exit(1);
                }
      
                XBuf_SHMInfo.shmid = shmget(IPC_PRIVATE, img->height * img->bytes_per_line, IPC_CREAT | 0777);
      
                if(XBuf_SHMInfo.shmid < 0)
                {
                        puts("FATAL DISPLAY ERROR: shmget\n\n"); fflush(stdout);
                        exit(1);
                }
      
                XBuf = img->data = XBuf_SHMInfo.shmaddr = shmat(XBuf_SHMInfo.shmid, 0, 0);
                if(!XBuf)
                {
                        puts("FATAL DISPLAY ERROR: shmat\n\n"); fflush(stdout);
                        exit(1);
                }

                XBuf_SHMInfo.readOnly = False;
      
                if(!XShmAttach(dsp, &XBuf_SHMInfo))
                {
                        puts("FATAL DISPLAY ERROR: XShmAttach\n\n"); fflush(stdout);
                        exit(1);
                }         

                shmctl(XBuf_SHMInfo.shmid, IPC_RMID, &s);   /* if someone kills me shm should be freed */
        }
        else
#endif
        {
                img = XCreateImage(dsp, DefaultVisualOfScreen(scr), DEPTH, ZPixmap, 0, XBuf, WIDTH, HEIGHT, DEPTH, 0);
                if(!img)
                { 
                        puts("FATAL DISPLAY ERROR: XCreateImage\n\n"); fflush(stdout);
                        exit(1);
                }

                XBuf = img->data = malloc(img->height * img->bytes_per_line);
                if(!XBuf)
                {
                        puts("FATAL DISPLAY ERROR: malloc\n\n"); fflush(stdout);
                        exit(1);
                }

        }

} /* static void AllocDisplayMem() */










#ifdef OPENGL

void SetVisualSize(int w, int h)
{
        XResizeWindow(dsp, win, w, h);
        glViewport(0, 0, w, h);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        //glOrtho( -1.0, 1.0, -1.0, 1.0, -1.0, 1.0 );
        glOrtho(0, width, height, 0, -1, 1);

} /* void SetVisualSize(int w, int h) */


void InitVisualRenderBuffer()
{
//puts("InitVisualRenderBuffer"); fflush(stdout);
        width = ((uWORD *)mem.vi_reg)[2];
        height = (width*3)>>2;
        XGetWindowAttributes(dsp, win, &win_attrib);
        WIDTH  = win_attrib.width;
        HEIGHT = win_attrib.height;
        stretch_x = (float)WIDTH/width;
        stretch_y = (float)HEIGHT/((width*3)>>2);

#ifndef JUST3D
        DrawScreenGL(width, height);
#endif
    /* from FiRES and Gil */
        glOrtho(0, width, height, 0, -1, 1);
        glViewport(0, 0, WIDTH, HEIGHT);

        /* set viewer */
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        //glFrustum(-4.0, 4.0, -2.704225, 2.74225, 4.0, 4096.0);

        /* position viewer */
        glMatrixMode(GL_MODELVIEW);

        glCullFace(GL_BACK);
        glFrontFace(GL_CCW);
        glDisable(GL_CULL_FACE);

        glDepthFunc(GL_LEQUAL);
        //glEnable(GL_DEPTH_TEST);

        //speed!
        //glShadeModel(GL_SMOOTH);
        //glShadeModel(GL_FLAT);
    /* (niki) next 2 functions are a speed problem on slow machines */
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    /* (niki)if we enable this here we cannot switch it off later. why the hell? */
        /*glEnable(GL_BLEND);*/

        glAlphaFunc(GL_GEQUAL, 0.07f);
        glDisable(GL_ALPHA_TEST);

        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    /* (niki) this caused a seg fault */
    /*
        glEnable(GL_TEXTURE_2D);
    */

        glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

   /* from FiRES and Gil end */
}



void SetVisualRenderBuffer(int format, int size, int width, char *address)
{
//puts("SetVisualRenderBuffer"); fflush(stdout);

#ifndef JUST3D
        if(rdp_valid_addr)
                PutBackPixmap();
        else
                rdp_valid_addr = 1;
#endif

    /* format and size cannot be supported in X (as far as i know)
       we store it here ... maybe we can simulate it in FlushVisualRenderBuffer */
        rdp_format  = format;
        rdp_size    = size;
        rdp_width   = width;
        rdp_address = address;


#ifndef JUST3D
        glXDestroyGLXPixmap(dsp, gl_pix);
        XFreePixmap(dsp, back_pix);
        puts("pixmap destroyed"); fflush(stdout);
        XFreeGC(dsp, back_gc);

        back_pix = XCreatePixmap(dsp, RootWindowOfScreen(scr), rdp_width, rdp_width*3/4, DEPTH);
        back_gc  = XCreateGC(dsp, back_pix, 0, 0);
        gl_pix = glXCreateGLXPixmapMESA(dsp, vis, back_pix, win_attrib.colormap);
#endif

}





void FlushVisualRenderBuffer()
{

//puts("FlushVisualRenderBuffer"); fflush(stdout);
#ifndef JUST3D
        if(rdp_valid_addr)
                PutBackPixmap();

        valid_clip_rect = 0;
#endif

#ifdef JUST3D
        glXSwapBuffers(dsp, win);
        glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
        glClearDepth(1.0);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
#else
        //XCopyArea(dsp, back_pix, win, back_gc, 0, 0, WIDTH, HEIGHT, 0, 0);
#endif
        XFlush(dsp);


} /* void FlushVisualRenderBuffer() */






void SetVisualClipRectangle(int ulx, int uly, int lrx, int lry)
{
        return;

        back_cliprect[0].x       = (short)ulx;
        back_cliprect[0].y       = (short)uly;
        back_cliprect[0].width   = (short)(lrx-ulx);
        back_cliprect[0].height  = (short)(lry-uly);
/*
        back_cliprect[0].x       = (short)0;
        back_cliprect[0].y       = (short)0;
        back_cliprect[0].width   = (short)321;
        back_cliprect[0].height  = (short)241;
*/

        /* we do it right after allocating the pixmap too
           because in most cases this is set before SETCOLORIMAGE */

#ifdef JUST3D
        back_cliprect[0].x      = (int)(float)back_cliprect[0].x      * stretch_x;
        back_cliprect[0].y      = (int)(float)back_cliprect[0].y      * stretch_y;
        back_cliprect[0].width  = (int)(float)back_cliprect[0].width  * stretch_x;
        back_cliprect[0].height = (int)(float)back_cliprect[0].height * stretch_y;
        XSetClipRectangles(dsp, gc, 0, 0, back_cliprect, 1, Unsorted);
#else
        if(valid_back_pix)
        {
                XSetClipRectangles(dsp, back_gc, 0, 0, back_cliprect, 1, Unsorted);
        }
        else
        {
                valid_clip_rect = 1;
        }
#endif

} /* void SetVisualClipRect() */

#endif /* #ifdef OPENGL */








/*****************************************************************************\
*
* everything below is coded in ../gl_renderer.c if OPENGL is used.
* otherwise the routines in this file are compiled.
*
* oh - the routines DrawVisualTexTriangle() and DrawVisualTexRectangle() are
* missing here. DrawVisualTriangle() and DrawVisualRectangle() are called
* instead, because it makes no sense (in my opinion) to do texturing without 3d
* lib.
*
* i know that this is not the best style ... but for now it works.
* - niki
*
\*****************************************************************************/





#ifndef OPENGL   /* otherwise it is in ../gl_renderer.c */





void SetVisualSize(int w, int h)
{
        XResizeWindow(dsp, win, w, h);

} /* void SetVisualSize(int w, int h) */





void InitVisualRenderBuffer()
{
//puts("InitVisualRenderBuffer"); fflush(stdout);
#ifdef JUST3D
        width = ((uWORD *)mem.vi_reg)[2];
        height = (width*3)>>2;
        XGetWindowAttributes(dsp, win, &win_attrib);
        WIDTH  = win_attrib.width;
        HEIGHT = win_attrib.height;
        stretch_x = (float)WIDTH/width;
        stretch_y = (float)HEIGHT/((width*3)>>2);

        XSetForeground(dsp, gc, (long)0x0);
        XFillRectangle(dsp, win, gc, 0, 0, WIDTH-1, HEIGHT-1);
#endif
}





void SetVisualRenderBuffer(int format, int size, int width, char *address)
{
#ifndef JUST3D
        if(rdp_valid_addr)
                PutBackPixmap();
        else
                rdp_valid_addr = 1;

    /* format and size cannot be supported in X (as far as i know)
       we store it here ... maybe we can simulate it in FlushVisualRenderBuffer */
        rdp_format  = format;
        rdp_size    = size;
        rdp_width   = width;
        rdp_address = address;

//puts("SetVisualRenderBuffer"); fflush(stdout);

        XFreePixmap(dsp, back_pix);
        puts("pixmap destroyed"); fflush(stdout);
        XFreeGC(dsp, back_gc);

        back_pix = XCreatePixmap(dsp, RootWindowOfScreen(scr), rdp_width, rdp_width*3/4, DEPTH);
        back_gc  = XCreateGC(dsp, back_pix, 0, 0);
    /* clear the pixmap */
        XSetForeground(dsp, back_gc, (long)0x0);
        XFillRectangle(dsp, back_pix, back_gc, 0, 0, rdp_width, rdp_width*3/4);
/*
        if(valid_clip_rect)
                XSetClipRectangles(dsp, back_gc, 0, 0, back_cliprect, 1, Unsorted);
*/
#endif
}





#ifndef JUST3D
static void PutBackPixmap()
{
        XImage  *src_img;
        void    *src;
        void    *dst;
        int     i;


        if(rdp_format == 0 && rdp_size == 2)
        {
                dst = rdp_address;

                src_img = XGetImage(dsp, back_pix, 0, 0, width, height, 0xffffffff, ZPixmap);

                src = src_img->data;

                for(i=0; i<((width*height)<<1); i++)
                {
#ifdef ENDIAN_IS_LITTLE
                        //FIX
                        if(((uBYTE *)src)[i & (~1)] & 0x01)   /* alpha stuff */
                        {
                                ((uBYTE *)dst)[i^3] = ((uBYTE *)src)[i];
                        }
#endif
#ifdef ENDIAN_IS_BIG
                        //FIX
                        //if(((uBYTE *)src)[i] & 0x01)   /* alpha stuff */
                        //{
                                ((uBYTE *)dst)[i] = ((uBYTE *)src)[i];
                        //}
#endif
                                
                }

                XDestroyImage(src_img);
        }

} /* static void PutBackPixmap() */
#endif





void FlushVisualRenderBuffer()
{
#ifndef JUST3D


//puts("FlushVisualRenderBuffer"); fflush(stdout);

        if(rdp_valid_addr)
                PutBackPixmap();

        valid_clip_rect = 0;
#endif

        XFlush(dsp);

} /* void FlushVisualRenderBuffer() */





void SetVisualClipRectangle(int ulx, int uly, int lrx, int lry)
{
        return;

        back_cliprect[0].x       = (short)ulx;
        back_cliprect[0].y       = (short)uly;
        back_cliprect[0].width   = (short)(lrx-ulx);
        back_cliprect[0].height  = (short)(lry-uly);
/*
        back_cliprect[0].x       = (short)0;
        back_cliprect[0].y       = (short)0;
        back_cliprect[0].width   = (short)321;
        back_cliprect[0].height  = (short)241;
*/

        /* we do it right after allocating the pixmap too
           because in most cases this is set before SETCOLORIMAGE */

#ifdef JUST3D
        back_cliprect[0].x      = (int)(float)back_cliprect[0].x      * stretch_x;
        back_cliprect[0].y      = (int)(float)back_cliprect[0].y      * stretch_y;
        back_cliprect[0].width  = (int)(float)back_cliprect[0].width  * stretch_x;
        back_cliprect[0].height = (int)(float)back_cliprect[0].height * stretch_y;
        XSetClipRectangles(dsp, gc, 0, 0, back_cliprect, 1, Unsorted);
#else
        if(valid_back_pix)
        {
                XSetClipRectangles(dsp, back_gc, 0, 0, back_cliprect, 1, Unsorted);
        }
        else
        {
                valid_clip_rect = 1;
        }
#endif

} /* void SetVisualClipRect() */





void SetVisualColor(int color)
{
#ifdef JUST3D
        XSetForeground(dsp, gc, (long)color);
#else
        XSetForeground(dsp, back_gc, (long)color);
#endif
}





void DrawVisualFillRect(int ulx, int uly, int lrx, int lry)
{
#ifdef JUST3D
        XFillRectangle(dsp, win, gc, ulx*stretch_x, uly*stretch_y, (lrx-ulx)*stretch_x, (lry-uly)*stretch_y);
#else
        XFillRectangle(dsp, back_pix, back_gc, ulx, uly, lrx-ulx, lry-uly);
#endif

} /* void DrawVisualFillRect() */





void ClearVisual()
{
#ifdef JUST3D
        XSetForeground(dsp, gc, (long)0);
        XFillRectangle(dsp, win, gc, 0, 0, WIDTH-1, HEIGHT-1);
#else
        XSetForeground(dsp, back_gc, (long)0);
        XFillRectangle(dsp, back_pix, back_gc, 0, 0, width-1, ((width*3)>>2)-1);
#endif

} /* void ClearVisual() */





void DrawVisualLine(int ulx, int uly, int lrx, int lry)
{
#ifdef JUST3D
        XDrawLine(dsp, win, gc, ulx*stretch_x, uly*stretch_y, lrx*stretch_x, lry*stretch_y);
#else
        XDrawLine(dsp, back_pix, back_gc, ulx, uly, lrx, lry);
#endif

} /* void DrawVisualLine() */





void DrawVisualTriangle(int x1, int y1, int x2, int y2, int x3, int y3)
{
        XPoint p[4];

        p[0].x = x1;
        p[0].y = y1;
        p[1].x = x2;
        p[1].y = y2;
        p[2].x = x3;
        p[2].y = y3;
        p[3].x = x1;
        p[3].y = y1;

#ifdef JUST3D
        width = ((uWORD *)mem.vi_reg)[2];
        if(width!=0)
        {
                p[0].x = (int)(float)p[0].x * stretch_x;
                p[0].y = (int)(float)p[0].y * stretch_y;
                p[1].x = (int)(float)p[1].x * stretch_x;
                p[1].y = (int)(float)p[1].y * stretch_y;
                p[2].x = (int)(float)p[2].x * stretch_x;
                p[2].y = (int)(float)p[2].y * stretch_y;
                p[3].x = p[0].x;
                p[3].y = p[0].y;
        }
        XDrawLines(dsp, win, gc, p, 4, CoordModeOrigin);
#else
        XDrawLines(dsp, back_pix, back_gc, p, 4, CoordModeOrigin);
#endif

} /* void DrawVisualTriangle() */





void DrawVisualRectangle(int ulx, int uly, int lrx, int lry)
{
#ifdef JUST3D
        XDrawRectangle(dsp, win, gc, ulx*stretch_x, uly*stretch_y, (lrx-ulx)*stretch_x, (lry-uly)*stretch_y);
#else
        XDrawRectangle(dsp, back_pix, back_gc, ulx, uly, lrx-ulx, lry-uly);
#endif

} /* void DrawVisualRectangle() */





#endif /* #ifndef OPENGL */

